#ifndef _TYPE_H
#define _TYPE_H

#include <stdlib.h>
#include <stdio.h>
#include "util.h"

typedef enum {
	T_BOOL,
	T_INT,
	T_REAL,
	T_MARK,
	T_NULL,
	T_OP,
	T_ARRAY,
	T_DICT,
	T_STRING,
	T_NAME,
	T_GSTATE
} type_t;

typedef enum {
	C_INVALID,
	C_LT,
	C_EQ,
	C_GT
} objcmp_t;

typedef struct {
	double *data;
	int size;
	int cnt;
} spath_t;

spath_t * spath_init();
void spath_add(spath_t *p, double x, double y);
spath_t * spath_copy(spath_t *p);
double *spath_pt(spath_t *p, int i);
double *spath_last(spath_t *p);

typedef struct {
	spath_t **sp;
	int cnt;
	int wasmoveto;
	int closed;
} path_t;

path_t * path_init();
path_t * path_copy(path_t *p);
spath_t * path_last(path_t *p);
void path_newsub(path_t *p);

typedef struct gstate gstate_t;

struct gstate {
	double CTM[9];
	path_t *path;
	path_t *clippath;
	int cr, cg, cb;
	double thick;
};

extern gstate_t cur_gstate;

typedef void (* op_t) (void);

typedef struct obj obj_t;

struct obj {
	type_t type;
	int exec;
	union {
		int i;
		double r;
		struct {
			int size;
			obj_t **v;
		} arr;
		char *s;
		op_t op;
		ptrhash_t *dict;
		gstate_t *gs;
	};
};

obj_t *mkobj(type_t type);
obj_t *mkint(int v);
obj_t *mkreal(double r);
obj_t *mkbool(int v);
obj_t *mkarr();
obj_t *mkstring(const char *s);
obj_t *mkname(const char *s);
obj_t *mkename(const char *s);
obj_t *mkop(op_t op);
obj_t *mknull();
obj_t *mkmark();
obj_t *mkdict();
obj_t *mkgstate(gstate_t *g);

int obj_hash_eq(obj_t *a, obj_t *b);
unsigned int obj_hash(obj_t *o);

objcmp_t obj_cmp(obj_t *a, obj_t *b);

obj_t * arr_idx(obj_t *arr, int i);
void arr_setidx(obj_t *arr, int i, obj_t *o);
void arr_add(obj_t *arr, obj_t *o);
obj_t * arr_pop(obj_t *arr);
obj_t * arr_peek(obj_t *arr);

obj_t * dict_lookup(obj_t *dict, obj_t *key);
void dict_set(obj_t *dict, obj_t *key, obj_t *val);

static inline double getreal(obj_t *o)
{
	switch(o->type) {
	case T_INT:
		return (double) o->i;
	case T_REAL:
		return o->r;
	default:
		fprintf(stderr, "cannot cast object to real\n");
		abort();
	}
}

#endif
