#ifndef _CMD_H
#define _CMD_H

#include "SDL.h"
#include "draw.h"

typedef struct {
	/* image in native format for fast blitting */
	SDL_Surface *native;

	/* image in 32bpp format for texturing */
	SDL_Surface *b32;

	/* whether we have an alpha channel */
	int alpha;
} image_t;

extern ptrhash_t *imagehash; /* image name -> image_t * hash */
extern Uint32 colormap[256]; /* colormap used for IMAGE_SET */

void image_init();
void image_load(char *symbol, char *path);
void image_create(char *symbol, int w, int h);
void set_colormap(int index, Uint32 color);
void image_update(char *symbol, int x, int y, char *string);

int color_lookup(Uint32 *out, char *name);
int color_hex(Uint32 *out, char *hex);

void set_color(Uint32 clr);
void set_bg(Uint32 clr);
void set_thick(float thick);
void set_alpha(Uint8 alpha);
void set_dash(unsigned int dash);

void pushmatrix();
void popmatrix();

void multmatrix(matrix_t *m);
void rotate(double angle);
void scale(double x, double y);
void shift(double x, double y);
void loadident();
void viewport(double left, double bottom, double width, double height);
void viewport_macro(char *name, int aspect);

void eventreport(int on);

void frame();
void flush();
void saveframeas(const char *name);
void dumpframe(const char *name);

void drawdot(double x, double y);
void drawline(double x1, double y1, double x2, double y2);
void drawarrow(double x1, double y1, double x2, double y2);
void drawrect(double cx, double cy, double w, double h);
void drawrect2(double x1, double y1, double x2, double y2);
void drawtext(double x, double y, int horalign, int veralign, const char *text);
void drawblit(double x, double y, int horalign, int veralign, char *image);
void drawlinefrom(double x, double y);
void drawlineto(double x, double y);
void drawpoly(polygon_t *p);
void drawcircle(double cx, double cy, double r, double verts);
void drawimgpoly(double ax, double ay, double ux, double uy, double vx, double vy, char *img, polygon_t *p);

void clip(double x, double y, double w, double h);

void cursor_init();
void defcursor(char *name, const char *data, const char *mask, int hot_x, int hot_y);
void setcursor(char *name);

void yyerror(const char *str, ...);
extern int yyline;

typedef struct matrixstack matrixstack_t;

struct matrixstack {
	matrix_t m;
	matrixstack_t *next;
};

extern matrixstack_t *matrixstack;

void transform(double x, double y, double *ox, double *oy);

extern int keepalive, headless;

int doinput(char *buf, int size);

void macro_execute(char *name);
void macro_boundingbox(char *name, int aspect, matrix_t *out_matrix);

void polybuffer_add(double x, double y);
polygon_t *polybuffer_end();
polygon_t *polygon_copy(polygon_t *p);

polygon_t *makerect(double cx, double cy, double w, double h);
polygon_t *makerect2(double x1, double y1, double x2, double y2);
polygon_t *makecircle(double cx, double cy, double r, double verts);

/* returns time since epoch in seconds */
double dbltime(void);

void flushdelay(double tm);

#endif
