#include <math.h>
#include <assert.h>
#include <stdlib.h>
#include "cmd.h"
#include "draw.h"
#include "gfx.h"
#include "util.h"
#include "area.h"

ptrhash_t *macrohash = NULL;

Uint32 background = 0xffffff;

drawcmd_t *drawarray = NULL;
size_t drawarray_alloc = 0;
size_t drawarray_cnt = 0;

void draw(const drawcmd_t *cmd)
{
	if(drawarray_cnt == drawarray_alloc) {
		if(!drawarray_alloc) drawarray_alloc = 128;
		else drawarray_alloc *= 2;
		drawarray = realloc(drawarray, drawarray_alloc * sizeof(*drawarray));
		if(!drawarray) abort();
	}

	drawarray[drawarray_cnt ++] = *cmd;
}

void resetdraw()
{
	int i;
	drawcmd_t *d;
	for(i=0, d=drawarray; i<drawarray_cnt; i++, d++) {
		switch(d->type) {
		case DTEXT:
		case DBLIT:
			free(d->text.str);
			break;
		case DPOLY:
			free(d->poly);
			break;
		case DIMGPOLY:
			free(d->imgpoly.img);
			free(d->imgpoly.poly);
			break;
		default:
			;
		}
	}

	drawarray_cnt = 0;
}

void beginframe()
{
	resetdraw();
}

int keepaspect = 1;

/* get text width/height (in chars) */
static void textdim(char *txt, int *w, int *h)
{
	int x = 0, y = 1, c = 0;
	char *p;
	for(p=txt; *p; p++) {
		if(*p == '\n') {
			c = 0;
			y ++;
			continue;
		}
		c ++;
		if(c > x) x = c;
	}
	*w = x;
	*h = y;
}

static double ox=0, oy=0, mx=640, my=-480;

void untransform(int x, int y, double *outx, double *outy)
{
	double *m = eventmatrix.v;

	double ix = ((double) x - ox) / mx;
	double iy = ((double) y - oy) / my;

	/*
	ix = outx * m[0] + outy * m[1] + m[2];
	iy = outx * m[3] + outy * m[4] + m[5];

	(ix - m[2] - outy * m[1]) / m[0] = outx;

	iy - m[5] - ((ix - m[2] - outy * m[1]) / m[0]) * m[3] = outy * m[4];

	iy - m[5] - (ix - m[2] - outy * m[1]) / m[0] * m[3] = outy * m[4]

	iy - m[5] - ix / m[0] * m[3] + m[2] / m[0] * m[3] + outy * m[1] / m[0] * m[3] = outy * m[4]

	iy - ix / m[0] * m[3] + m[2] / m[0] * m[3] - m[5] = outy * m[4] - outy * m[1] / m[0] * m[3]

	iy - ix / m[0] * m[3] + m[2] / m[0] * m[3] - m[5] = outy * (m[4] - m[1] / m[0] * m[3])

	(iy - ix / m[0] * m[3] + m[2] / m[0] * m[3] - m[5]) / (m[4] - m[1] / m[0] * m[3]) = outy

	..

	*/

	*outx = 0;
	*outy = 0;

	if(m[0] == 0) return;
	double m03 = m[3] / m[0];
	double z = m[4] - m[1] * m03;
	if(z == 0) return;

	*outy = (iy - ix * m03 + m[2] * m03 - m[5]) / z;
	*outx = (ix - m[2] - (*outy) * m[1]) / m[0];
}

void render()
{
	gfxclr *fgcolor;
	Uint8 alpha = 255;
	gfxclr defaultfgcolor = mkclr(0);
	float thick = 1.0;
	fgcolor = &defaultfgcolor;
	static int *polypts = NULL;
	static int polycnt = 0;
	unsigned int dash = 0xffff;
	dashptr = 0;

	ox = main_surface->w/2;
	oy = main_surface->h/2;
	if(keepaspect) {
		if(main_surface->w < main_surface->h) {
			mx = main_surface->w / 2;
			my = - main_surface->w / 2;
		} else {
			mx = main_surface->h / 2;
			my = - main_surface->h / 2;
		}
	} else {
		mx = main_surface->w / 2;
		my = - main_surface->h / 2;
	}

	areamap_ox = ox / 2;
	areamap_oy = oy / 2;
	areamap_mx = mx / 2;
	areamap_my = my / 2;

	gfxclr bg = mkclr(background);

	main_surface->clip_rect.x = 0;
	main_surface->clip_rect.y = 0;
	main_surface->clip_rect.w = main_surface->w;
	main_surface->clip_rect.h = main_surface->h;

	if(!noclrscr)
		SDL_FillRect(main_surface, NULL, CLR(bg));

	int i,j,k;
	drawcmd_t *d;
	for(i=0, d=drawarray; i<drawarray_cnt; i++, d++) {
		switch(d->type) {
		case DCOLOR:
			fgcolor = &d->color;
			break;
		case DALPHA:
			alpha = d->alpha;
			break;
		case DDASH:
			dash = d->dash;
			break;
		case DDOT:
			if(alpha == 255)
				gfx_dot(main_surface, ox+d->dot.x*mx, oy+d->dot.y*my, CLR(*fgcolor));
			else
				gfx_dot_alpha(main_surface, ox+d->dot.x*mx, oy+d->dot.y*my, alpha, CLR(*fgcolor));
			break;
		case DTHICK:
			thick = d->thick;
			break;
		case DLINE:
			if((int)(thick+0.5) == 1) {
				if(dash == 0xffff) {
					if(alpha == 255)
						gfx_line(main_surface, ox+d->line.x1*mx, oy+d->line.y1*my,
								 ox+d->line.x2*mx, oy+d->line.y2*my,
								 CLR(*fgcolor));
					else
						gfx_line_alpha(main_surface, ox+d->line.x1*mx, oy+d->line.y1*my,
									   ox+d->line.x2*mx, oy+d->line.y2*my,
									   alpha, CLR(*fgcolor));
				} else {
					if(alpha == 255)
						gfx_line_dash(main_surface, ox+d->line.x1*mx, oy+d->line.y1*my,
								 ox+d->line.x2*mx, oy+d->line.y2*my,
								 CLR(*fgcolor), dash);
					else
						gfx_line_alpha_dash(main_surface, ox+d->line.x1*mx, oy+d->line.y1*my,
										   ox+d->line.x2*mx, oy+d->line.y2*my,
										   alpha, CLR(*fgcolor), dash);
				}
			} else {
				if(dash == 0xffff)
					d_line_thick(ox+d->line.x1*mx, oy+d->line.y1*my,
								 ox+d->line.x2*mx, oy+d->line.y2*my,
								 thick, alpha, CLR(*fgcolor));
				else
					d_line_thick_dash(ox+d->line.x1*mx, oy+d->line.y1*my,
									  ox+d->line.x2*mx, oy+d->line.y2*my,
									  thick, alpha, CLR(*fgcolor), dash);
			}
			break;
		case DTEXT:
			{
				int w, h;
				textdim(d->text.str, &w, &h);
				w *= font_w;
				h *= font_h;
				int x = ox+d->text.x*mx;
				int y = oy+d->text.y*my;
				switch(d->text.horalign) {
				case -1:
					x -= w;
					break;
				case 0:
					x -= (w / 2);
					break;
				}
				switch(d->text.veralign) {
				case 1:
					y -= h;
					break;
				case 0:
					y -= (h / 2);
					break;
				}
				d_ml_text(x, y, d->text.str, alpha, CLR(*fgcolor));
			}
			break;
		case DBLIT:
			{
				ptrhash_entry_t *imge = ptrhash_getentry(imagehash, d->text.str);
				if(!imge) {
					fprintf(stderr, "blit cannot find image: %s\n", d->text.str);
					exit(1);
				}
				image_t *img = imge->value;

				int w = img->native->w;
				int h = img->native->h;

				int x = ox+d->text.x*mx;
				int y = oy+d->text.y*my;
				switch(d->text.horalign) {
				case -1:
					x -= w;
					break;
				case 0:
					x -= (w / 2);
					break;
				}
				switch(d->text.veralign) {
				case 1:
					y -= h;
					break;
				case 0:
					y -= (h / 2);
					break;
				}

				SDL_Rect dst = { x, y, w, h };

				if(alpha == 255) {
					SDL_BlitSurface(img->native, NULL, main_surface, &dst);
				} else {
					if(img->alpha) {
						/* img has alpha channel - have to multiply */
						gfx_multalpha_blit(img->b32, main_surface, x, y, alpha);
					} else {
						/* no own alpha channel, can set surface alpha */
						SDL_SetAlpha(img->b32, SDL_SRCALPHA, alpha);
						SDL_BlitSurface(img->b32, NULL, main_surface, &dst);
						SDL_SetAlpha(img->b32, 0, 255);
					}
				}
			}
			break;
		case DPOLY:
			if(polycnt < d->poly->npoints) {
				polypts = realloc(polypts, d->poly->npoints * 2 * sizeof(int));
				if(!polypts) abort();
				polycnt = d->poly->npoints;
			}
			if(d->poly->aa) {
				/* "anti-aliasing" mode */
				for(j=0,k=0; j<d->poly->npoints; j++,k+=2) {
					polypts[k]   = (ox+d->poly->points[k]*mx) * (float) POLY_A_XMULT;
					polypts[k+1] = (oy+d->poly->points[k+1]*my) * (float) POLY_A_YMULT;
				}
				gfx_fillpoly_aa(main_surface, d->poly->npoints, (gfx_polypt) polypts, alpha, CLR(*fgcolor));
			} else {
				/* normal mode */
				for(j=0,k=0; j<d->poly->npoints; j++,k+=2) {
					polypts[k]   = ox+d->poly->points[k]*mx;
					polypts[k+1] = oy+d->poly->points[k+1]*my;
				}
				gfx_fillpoly(main_surface, d->poly->npoints, (gfx_polypt) polypts, alpha, CLR(*fgcolor));
			}
			break;
		case DIMGPOLY:
			{
				ptrhash_entry_t *imge = ptrhash_getentry(imagehash, d->imgpoly.img);
				if(!imge) {
					fprintf(stderr, "imgpoly cannot find image: %s\n", d->imgpoly.img);
					exit(1);
				}
				image_t *img = imge->value;

				if(polycnt < d->imgpoly.poly->npoints) {
					polypts = realloc(polypts, d->imgpoly.poly->npoints * 2 * sizeof(int));
					if(!polypts) abort();
					polycnt = d->imgpoly.poly->npoints;
				}

				for(j=0,k=0; j<d->imgpoly.poly->npoints; j++,k+=2) {
					polypts[k]   = ox+d->imgpoly.poly->points[k]*mx;
					polypts[k+1] = oy+d->imgpoly.poly->points[k+1]*my;
				}

				double ax = ox + d->imgpoly.ax * mx;
				double ay = oy + d->imgpoly.ay * my;
				double ux = ox + d->imgpoly.ux * mx;
				double uy = oy + d->imgpoly.uy * my;
				double vx = ox + d->imgpoly.vx * mx;
				double vy = oy + d->imgpoly.vy * my;

				ux -= ax;
				uy -= ay;
				vx -= ax;
				vy -= ay;

				double A = 1.0 / (ux * vy - vx * uy);
				double xu = A * vy;
				double yu = - A * vx;
				double xv = - A * uy;
				double yv = A * ux;

				//printf("ux=%f uy=%f vx=%f vy=%f xu=%f xv=%f yu=%f yv=%f\n", ux, uy, vx, vy, xu, xv, yu, yv);

				gfx_imgpoly(main_surface, d->imgpoly.poly->npoints, (gfx_polypt) polypts, alpha, ax, ay, xu, xv, yu, yv, img->alpha, img->alpha ? img->b32 : img->native);
			}

			break;

		case DCLIP:
			{
				int sx1 = (int) (ox + d->line.x1 * mx);
				int sy1 = (int) (oy + d->line.y1 * my);
				int sx2 = (int) (ox + d->line.x2 * mx);
				int sy2 = (int) (oy + d->line.y2 * my);
				int t;

				if(sx1 > sx2) {
					t = sx1;
					sx1 = sx2;
					sx2 = t;
				}

				if(sy1 > sy2) {
					t = sy1;
					sy1 = sy2;
					sy2 = t;
				}

				if(sx1<0) sx1=0;
				if(sy1<0) sy1=0;
				if(sx2 >= main_surface->w) sx2 = main_surface->w - 1;
				if(sy2 >= main_surface->h) sy2 = main_surface->h - 1;

				main_surface->clip_rect.x = sx1;
				main_surface->clip_rect.y = sy1;
				main_surface->clip_rect.w = sx2 - sx1 + 1;
				main_surface->clip_rect.h = sy2 - sy1 + 1;
			}
			break;
		default:;
		}
	}
}

void warpmouse(double x, double y)
{
	double xx, yy;
	transform(x, y, &xx, &yy);
	SDL_WarpMouse(ox+xx*mx, oy+yy*my);
}

void transform_pixel(double x, double y, int *outx, int *outy)
{
	double xx, yy;
	transform(x, y, &xx, &yy);
	*outx = (int) (ox + xx * mx);
	*outy = (int) (oy + yy * my);
}

/* initialize macros */
void init_macros()
{
	macrohash = strhash_new();
}

/* free macro */
void free_macro(macro_t *m)
{
	int i;
	macro_el_t *el;
	for(i=0, el=m->els; i<m->cnt; i++, el++) {
		switch(el->type) {
		case MTEXT:
		case MBLIT:
			free(el->text.str);
			break;
		case MIMGPOLY:
			free(el->imgpoly.img);
			free(el->imgpoly.poly);
			break;
		case MINVOKE:
			free(el->invoke);
			break;
		case MMACROVIEWPORT:
			free(el->viewport_macro.name);
			break;
		case MPOLY:
			free(el->poly);
			break;
		case MAREA:
			free(el->area.name);
			free(el->area.poly);
			break;
		case MDELAREA:
			free(el->area.name);
			break;
		case MIMAGESET:
			free(el->image_set.img);
			free(el->image_set.string);
		default:;
		}
	}

	free(m->els);

	free(m);
}

/* current macro */
static int curmacro_alloc = 0;
static macro_el_t *curmacro_els = NULL;
static int curmacro_cnt = 0;
static char *curmacro_name = NULL;

/* begin recording macro */
void begin_macro(const char *name, int append)
{
	assert(!curmacro_name && !curmacro_els && !curmacro_cnt && !curmacro_alloc);

	if(append) {
		ptrhash_entry_t *ent = ptrhash_getentry(macrohash, (char *) name);
		if(ent) {
			macro_t *m = ent->value;
			curmacro_alloc = m->cnt;
			curmacro_cnt = m->cnt;
			curmacro_els = m->els;
		}
	}

	curmacro_name = strdup(name);
}

static void macro_append(macro_el_t *el)
{
	if(curmacro_cnt == curmacro_alloc) {
		if(!curmacro_alloc) curmacro_alloc = 32; else curmacro_alloc *= 2;
		curmacro_els = realloc(curmacro_els, curmacro_alloc * sizeof *curmacro_els);
		if(!curmacro_els) abort();
	}

	curmacro_els[curmacro_cnt ++] = *el;
}

void macro_dot(double x, double y)
{
	macro_el_t c;
	c.type = MDOT;
	c.dot.x = x;
	c.dot.y = y;
	macro_append(&c);
}

void macro_color(Uint32 clr)
{
	macro_el_t c;
	c.type = MCOLOR;
	c.color = mkclr(clr);
	macro_append(&c);
}

void macro_dash(unsigned int dash)
{
	macro_el_t c;
	c.type = MDASH;
	c.dash = dash;
	macro_append(&c);
}

void macro_alpha(Uint8 alpha)
{
	macro_el_t c;
	c.type = MALPHA;
	c.alpha = alpha;
	macro_append(&c);
}

void macro_set_thick(float thick)
{
	macro_el_t c;
	c.type = MTHICK;
	c.thick = thick;
	macro_append(&c);
}

void macro_line(double x1, double y1, double x2, double y2)
{
	macro_el_t c;
	c.type = MLINE;
	c.line.x1 = x1;
	c.line.x2 = x2;
	c.line.y1 = y1;
	c.line.y2 = y2;
	macro_append(&c);
}

void macro_arrow(double x1, double y1, double x2, double y2)
{
	macro_el_t c;
	c.type = MARROW;
	c.line.x1 = x1;
	c.line.x2 = x2;
	c.line.y1 = y1;
	c.line.y2 = y2;
	macro_append(&c);
}

void macro_rect(double cx, double cy, double w, double h)
{
	macro_rect2(cx, cy, cx+w, cy+h);
}

void macro_rect2(double x1, double y1, double x2, double y2)
{
	macro_line(x1, y1, x2, y1);
	macro_line(x2, y1, x2, y2);
	macro_line(x2, y2, x1, y2);
	macro_line(x1, y2, x1, y1);
}

static double macrolinex = 0, macroliney = 0;
void macrolinefrom(double x, double y)
{
	macrolinex = x;
	macroliney = y;
}

void macrolineto(double x, double y)
{
	macro_line(macrolinex, macroliney, x, y);
	macrolinex = x;
	macroliney = y;
}

void macro_text(double x, double y, signed char horalign, signed char veralign, const char *str)
{
	macro_el_t c;
	c.type = MTEXT;
	c.text.x = x;
	c.text.y = y;
	c.text.horalign = horalign;
	c.text.veralign = veralign;
	c.text.str = strdup(str);
	macro_append(&c);
}

void macro_blit(double x, double y, signed char horalign, signed char veralign, const char *str)
{
	macro_el_t c;
	c.type = MBLIT;
	c.text.x = x;
	c.text.y = y;
	c.text.horalign = horalign;
	c.text.veralign = veralign;
	c.text.str = strdup(str);
	macro_append(&c);
}

void macro_imgpoly(double ax, double ay, double ux, double uy, double vx, double vy, char *img, polygon_t *p)
{
	macro_el_t c;
	c.type = MIMGPOLY;
	c.imgpoly.img = img;
	c.imgpoly.poly = p;
	c.imgpoly.ax = ax;
	c.imgpoly.ay = ay;
	c.imgpoly.ux = ux;
	c.imgpoly.uy = uy;
	c.imgpoly.vx = vx;
	c.imgpoly.vy = vy;
	macro_append(&c);
}

void macro_poly(polygon_t *p)
{
	macro_el_t c;
	c.type = MPOLY;
	c.poly = p;
	macro_append(&c);
}

void macro_push()
{
	macro_el_t c;
	c.type = MPUSH;
	macro_append(&c);
}

void macro_pop()
{
	macro_el_t c;
	c.type = MPOP;
	macro_append(&c);
}

void macro_mult(matrix_t *m)
{
	macro_el_t c;
	c.type = MMULT;
	c.matrix = *m;
	macro_append(&c);
}

#define PI_180 .0174532925199432957692369

void macro_rotate(double angle)
{
	int q;
	double a = remquo(angle, 90.0, &q) * PI_180;
	double ca = cos(a);
	double sa = sin(a);
	matrix_t m;

	switch (q&3) {
	case 0:
		m = (matrix_t){ .v = {
			ca, -sa, 0,
			sa, ca, 0,
			0, 0, 1.0
		}};
		break;
	case 1:
		m = (matrix_t){ .v = {
			-sa, -ca, 0,
			ca, -sa, 0,
			0, 0, 1.0
		}};
		break;
	case 2:
		m = (matrix_t){ .v = {
			-ca, sa, 0,
			-sa, -ca, 0,
			0, 0, 1.0
		}};
		break;
	case 3:
		m = (matrix_t){ .v = {
			sa, ca, 0,
			-ca, sa, 0,
			0, 0, 1.0
		}};
		break;
	default:
		abort();
	}

	macro_mult(&m);
}

void macro_scale(double x, double y)
{
	matrix_t m = (matrix_t) {
		.v = {
			x, 0, 0,
			0, y, 0,
			0, 0, 1.0
		}
	};
	macro_mult(&m);
}

void macro_shift(double x, double y)
{
	matrix_t m = (matrix_t) {
		.v = {
			1.0, 0, x,
			0, 1.0, y,
			0, 0, 1.0
		}
	};
	macro_mult(&m);
}

void macro_loadident()
{
	macro_el_t c;
	c.type = MIDENT;
	macro_append(&c);
}

void macro_viewport(double left, double bottom, double width, double height)
{
	/* shift -1 -1 * scale 2/width 2/height * shift -left -bottom */
	double w = 2.0/width;
	double h = 2.0/height;
	matrix_t m = (matrix_t) {
		.v = {
			w, 0, -1.0 - left*w,
			0, h, -1.0 - bottom*h,
			0, 0, 1
		}
	};

	macro_mult(&m);
}

void macro_viewport_macro(char* name, int aspect)
{
	macro_el_t c;
	c.type = MMACROVIEWPORT;
	c.viewport_macro.name = strdup(name);
	c.viewport_macro.aspect = aspect;
	macro_append(&c);
}

void macro_clip(double left, double bottom, double width, double height)
{
	macro_el_t c;
	c.type = MCLIP;
	c.clip.x = left;
	c.clip.y = bottom;
	c.clip.w = width;
	c.clip.h = height;
	macro_append(&c);
}

void macro_invokemacro(const char *name)
{
	macro_el_t c;
	c.type = MINVOKE;
	c.invoke = strdup(name);
	macro_append(&c);
}

void macro_circle(double cx, double cy, double r, double verts)
{
	int v = (int) verts;
	if(v < 3) v = 3;
	double step = 3.14159265358979 * 2.0 / (double) v;
	double a = step;
	double fx = r + cx;
	double fy = cy;
	double ox = fx;
	double oy = fy;
	int i;
	for(i=1; i<v; i++) {
		double x = cos(a) * r + cx;
		double y = sin(a) * r + cy;
		macro_line(ox, oy, x, y);
		ox = x;
		oy = y;
		a += step;
	}
	macro_line(ox, oy, fx, fy);
}

void macro_area(char *name, polygon_t *p)
{
	macro_el_t c;
	c.type = MAREA;
	c.area.name = name;
	c.area.poly = p;
	macro_append(&c);
}

void macro_delarea(char *name)
{
	macro_el_t c;
	c.type = MDELAREA;
	c.area.name = name;
	macro_append(&c);
}

void macro_colormap(int index, Uint32 color)
{
	macro_el_t c;
	c.type = MCOLORMAP;
	c.colormap.index = index;
	c.colormap.color = color;
	macro_append(&c);
}

void macro_image_update(char *image, int x, int y, char *string)
{
	/* NB: string ownership is taken!!! */
	macro_el_t c;
	c.type = MIMAGESET;
	c.image_set.img = image;
	c.image_set.x = x;
	c.image_set.y = y;
	c.image_set.string = string;
	macro_append(&c);
}

/* end recording macro */
void end_macro(int append)
{
	if(curmacro_cnt != curmacro_alloc) {
		curmacro_els = realloc(curmacro_els, curmacro_alloc * sizeof *curmacro_els);
		if(!curmacro_els) abort();
	}

	ptrhash_entry_t *prevmacro = ptrhash_getentry(macrohash, curmacro_name);

	if(prevmacro) {
		if(append) {
			/* appending to existing macro */

			macro_t *m = prevmacro->value;
			m->cnt = curmacro_cnt;
			m->els = curmacro_els;

			free(curmacro_name);
		} else {
			/* replacing existing macro */

			char *oldkey = prevmacro->key;
			prevmacro->key = curmacro_name;
			free(oldkey);

			free_macro(prevmacro->value);

			macro_t *m = malloc(sizeof *m);
			if(!m) abort();
			m->cnt = curmacro_cnt;
			m->els = curmacro_els;

			prevmacro->value = m;
		}
	} else {
		/* new macro */

		macro_t *m = malloc(sizeof *m);
		if(!m) abort();
		m->cnt = curmacro_cnt;
		m->els = curmacro_els;

		ptrhash_set(macrohash, curmacro_name, m);
	}

	curmacro_name = NULL;
	curmacro_els = NULL;
	curmacro_cnt = 0;
	curmacro_alloc = 0;
}

/* kill all macros */
void kill_macros()
{
	ptrhash_entry_t *e;
	for(e = ptrhash_first(macrohash); e; e = ptrhash_next(macrohash, e)) {
		free_macro(e->value);
		free(e->key);
	}
	ptrhash_clear(macrohash);
}

static void boundingbox_include(double *x1, double *y1, double *x2, double *y2, double x, double y)
{
	double ox = 0.0, oy = 0.0;
	transform(x, y, &ox, &oy);
	if(ox < *x1) *x1 = ox;
	if(ox > *x2) *x2 = ox;
	if(oy < *y1) *y1 = oy;
	if(oy > *y2) *y2 = oy;
}

void macro_boundingbox_expand(char* name, double *xmin, double *ymin, double *xmax, double *ymax)
{
	ptrhash_entry_t *entry = ptrhash_getentry(macrohash, name);
	if(!entry) {
		fprintf(stderr, "macro \"%s\" not found\n", name);
		exit(1);
	}

	macro_t *m = entry->value;

	int i, j;
	macro_el_t *el;
	for(i=0, el=m->els; i<m->cnt; i++, el++) {
		switch(el->type) {

		case MTEXT:
			// XXX: text size is in pixels
			boundingbox_include(xmin, ymin, xmax, ymax, el->text.x, el->text.y);
			break;

		case MBLIT:
			// XXX: image size is in pixels
			boundingbox_include(xmin, ymin, xmax, ymax, el->text.x, el->text.y);
			break;

		case MDOT:
			boundingbox_include(xmin, ymin, xmax, ymax, el->dot.x, el->dot.y);
			break;
		case MLINE:
			boundingbox_include(xmin, ymin, xmax, ymax, el->line.x1, el->line.y1);
			boundingbox_include(xmin, ymin, xmax, ymax, el->line.x2, el->line.y2);
			break;
		case MARROW:
			// XXX: doesn't consider arrowhead
			boundingbox_include(xmin, ymin, xmax, ymax, el->line.x1, el->line.y1);
			boundingbox_include(xmin, ymin, xmax, ymax, el->line.x2, el->line.y2);
			break;

		case MPOLY:
			for(j=0; j<el->poly->npoints; j+=2)
				boundingbox_include(xmin, ymin, xmax, ymax,
						el->poly->points[j], el->poly->points[j+1]);
			break;

		case MIMGPOLY:
			for(j=0; j<el->imgpoly.poly->npoints; j+=2)
				boundingbox_include(xmin, ymin, xmax, ymax,
						el->imgpoly.poly->points[j], el->imgpoly.poly->points[j+1]);
			break;

		case MPUSH:
			pushmatrix();
			break;
		case MPOP:
			popmatrix();
			break;
		case MMULT:
			multmatrix(&el->matrix);
			break;
		case MIDENT:
			loadident();
			break;

		case MINVOKE:
			macro_boundingbox_expand(el->invoke, xmin, ymin, xmax, ymax);
			break;

		default:;
		}
	}
}
