#ifndef _DRAW_H
#define _DRAW_H

#include "SDL.h"
#include "gfx.h"
#include "util.h"

extern ptrhash_t *macrohash;

typedef struct {
	double v[9];
} matrix_t;

typedef struct {
	short npoints;
	short aa;
	double points[];
} polygon_t;

typedef enum {
	DDOT,
	DLINE,
	DTEXT,
	DCOLOR,
	DALPHA,
	DTHICK,
	DDASH,
	DPOLY,
	DBLIT,
	DIMGPOLY,
	DCLIP
} drawtype_t;

typedef struct {
	drawtype_t type;
	union {
		gfxclr color;
		Uint8 alpha;
		float thick;
		unsigned int dash;

		struct {
			double x, y;
		} dot;

		struct {
			double x1, y1;
			double x2, y2;
		} line;

		struct {
			double x, y;
			signed char horalign, veralign;
			char *str;
		} text;

		struct {
			double ax, ay;
			double ux, uy;
			double vx, vy;
			char *img;
			polygon_t *poly;
		} imgpoly;

		polygon_t *poly;
	};
} drawcmd_t;

extern Uint32 background;

void draw(const drawcmd_t *cmd);
void resetdraw();
void untransform(int x, int y, double *outx, double *outy);

void beginframe();
void flushframe(int command);
void render();
void resize(int w, int h);
void warpmouse(double x, double y);

void transform_pixel(double x, double y, int *outx, int *outy);

/* saved frame count */
extern int saveframecount;
/* file name header for saved frames */
extern const char *saveframehead;

extern int eventreporting;
extern int keepaspect;
extern int noclrscr;

extern matrix_t eventmatrix;

typedef enum {
	MCOLOR,
	MDASH,
	MALPHA,
	MTHICK,
	MDOT,
	MLINE,
	MARROW,
	MTEXT,
	MBLIT,
	MPUSH,
	MPOP,
	MMULT,
	MIDENT,
	MINVOKE,
	MMACROVIEWPORT,
	MPOLY,
	MAREA,
	MDELAREA,
	MIMGPOLY,
	MCLIP,
	MCOLORMAP,
	MIMAGESET
} macro_el_type_t;

typedef struct {
	macro_el_type_t type;
	union {
		gfxclr color;
		Uint8 alpha;
		unsigned int dash;

		struct {
			double x, y;
		} dot;

		struct {
			double x1, y1;
			double x2, y2;
		} line;

		struct {
			double x, y;
			double w, h;
		} clip;

		struct {
			double x, y;
			signed char horalign, veralign;
			char *str;
		} text;

		struct {
			char *name;
			polygon_t *poly;
		} area;

		struct {
			double ax, ay;
			double ux, uy;
			double vx, vy;
			char *img;
			polygon_t *poly;
		} imgpoly;

		struct {
			int index;
			Uint32 color;
		} colormap;

		struct {
			char *img;
			int x, y;
			char *string;
		} image_set;

		struct {
			char *name;
			int aspect;
		} viewport_macro;

		matrix_t matrix;
		char *invoke;
		polygon_t *poly;
		float thick;
	};
} macro_el_t;

typedef struct {
	int cnt;
	macro_el_t *els;
} macro_t;

/* initialize macros */
void init_macros();

/* begin recording macro */
void begin_macro(const char *name, int append);

/* end recording macro */
void end_macro(int append);

/* kill all macros */
void kill_macros();

void macro_dot(double x, double y);
void macro_color(Uint32 clr);
void macro_dash(unsigned int dash);
void macro_alpha(Uint8 alpha);
void macro_set_thick(float thick);
void macro_line(double x1, double y1, double x2, double y2);
void macro_arrow(double x1, double y1, double x2, double y2);
void macro_text(double x, double y, signed char horalign, signed char veralign, const char *str);
void macro_blit(double x, double y, signed char horalign, signed char veralign, const char *str);
void macro_rect(double cx, double cy, double w, double h);
void macro_rect2(double x1, double y1, double x2, double y2);
void macro_push();
void macro_pop();
void macro_mult(matrix_t *m);
void macro_poly(polygon_t *p);
void macro_invokemacro(const char *name);
void macro_rotate(double angle);
void macro_scale(double x, double y);
void macro_shift(double x, double y);
void macro_loadident();
void macro_viewport(double left, double bottom, double width, double height);
void macro_viewport_macro(char* name, int aspect);
void macro_clip(double left, double bottom, double width, double height);
void macrolinefrom(double x, double y);
void macrolineto(double x, double y);
void macro_circle(double cx, double cy, double r, double verts);
void macro_area(char *name, polygon_t *p);
void macro_delarea(char *name);
void macro_imgpoly(double ax, double ay, double ux, double uy, double vx, double vy, char *img, polygon_t *p);
void macro_colormap(int index, Uint32 color);
void macro_image_update(char *image, int x, int y, char *string);
void macro_boundingbox_expand(char* name, double *xmin, double *ymin, double *xmax, double *ymax);

#endif
