#include "gfx.h"
#include <assert.h>

Uint32 gfx_colorsupdate_serial = 1;

SDL_Surface *main_surface = NULL;

unsigned int dashptr = 0x8000;
static inline void dashptr_advance()
{
	dashptr >>= 1;
	if(!dashptr) dashptr = 0x8000;
}

/* vertical line implementation 8bpp */
inline static void gfx_vline_8(SDL_Surface *s, int x, int y, int y2, Uint8 color)
{
	Uint8 *p = ((Uint8 *)s->pixels) + s->pitch * y + x;
	Uint8 *p2 = ((Uint8 *)s->pixels) + s->pitch * y2 + x;
	while(p<=p2) {
		*p = color;
		p += s->pitch;
	}
}

/* vertical line implementation 8bpp alpha */
inline static void gfx_vline_alpha_8(SDL_Surface *s, int x, int y, int y2, Uint8 alpha, Uint8 color)
{
	Uint8 *p = ((Uint8 *)s->pixels) + s->pitch * y + x;
	Uint8 *p2 = ((Uint8 *)s->pixels) + s->pitch * y2 + x;

	Uint32 ma = (alpha + 3) >> 5;
	Uint32 cg = (color & 0x1c) >> 2;
	Uint32 cng = color & 0xe3;
	Uint32 t1, t2;

	while(p<=p2) {
		t1 = *p & 0xe3;
		t2 = *p & 0x1c;
		*p = (((((cg - (t2 >> 2)) * ma) >> 1) + t2) & 0x1c) | (((((cng - t1) * ma) >> 3) + t1) & 0xe3);
		p += s->pitch;
	}
}

/* vertical line implementation 16bpp */
inline static void gfx_vline_16(SDL_Surface *s, int x, int y, int y2, Uint16 color)
{
	Uint8 *p = ((Uint8 *)s->pixels) + s->pitch * y + x*2;
	Uint8 *p2 = ((Uint8 *)s->pixels) + s->pitch * y2 + x*2;
	while(p<=p2) {
		*((Uint16 *)p) = color;
		p += s->pitch;
	}
}

/* vertical line implementation 15bpp w/ alpha */
inline static void gfx_vline_alpha_15(SDL_Surface *s, int x, int y, int y2, Uint8 alpha, Uint16 color)
{
	Uint8 *p = ((Uint8 *)s->pixels) + s->pitch * y + x*2;
	Uint8 *p2 = ((Uint8 *)s->pixels) + s->pitch * y2 + x*2;
	Uint32 ma = (alpha + 7) >> 3;
	Uint32 cg = (color & 0x03e0) >> 5;
	Uint32 cng = color & 0x7c1f;
	Uint32 t1, t2;

	while(p<=p2) {
		t1 = *((Uint16*)p) & 0x7c1f;
		t2 = *((Uint16*)p) & 0x03e0;
		*((Uint16*)p) = ((((cg - (t2 >> 5)) * ma) + t2) & 0x03e0) | (((((cng - t1) * ma) >> 5) + t1) & 0x7c1f);
		p += s->pitch;
	}
}

/* vertical line implementation 16bpp w/ alpha */
inline static void gfx_vline_alpha_16(SDL_Surface *s, int x, int y, int y2, Uint8 alpha, Uint16 color)
{
	Uint8 *p = ((Uint8 *)s->pixels) + s->pitch * y + x*2;
	Uint8 *p2 = ((Uint8 *)s->pixels) + s->pitch * y2 + x*2;
	Uint32 ma = (alpha + 7) >> 3;
	Uint32 cg = (color & 0x07e0) >> 5;
	Uint32 cng = color & 0xf81f;
	Uint32 t1, t2;

	while(p<=p2) {
		t1 = *((Uint16*)p) & 0xf81f;
		t2 = *((Uint16*)p) & 0x07e0;
		*((Uint16*)p) = ((((cg - (t2 >> 5)) * ma) + t2) & 0x07e0) | (((((cng - t1) * ma) >> 5) + t1) & 0xf81f);
		p += s->pitch;
	}
}

/* vertical line implementation 24bpp */
inline static void gfx_vline_24(SDL_Surface *s, int x, int y, int y2, Uint32 color)
{
	Uint8 *p = ((Uint8 *)s->pixels) + s->pitch * y + x*3;
	Uint8 *p2 = ((Uint8 *)s->pixels) + s->pitch * y2 + x*3;
	Uint8 c[3];
	if(SDL_BYTEORDER == SDL_BIG_ENDIAN) {
		c[0] = (color >> 16) & 0xff;
		c[1] = (color >> 8) & 0xff;
		c[2] = color & 0xff;
	} else {
		c[0] = color & 0xff;
		c[1] = (color >> 8) & 0xff;
		c[2] = (color >> 16) & 0xff;
	}

	while(p<=p2) {
		*p = c[0];
		*(p+1) = c[1];
		*(p+2) = c[2];

		p += s->pitch;
	}
}

/* vertical line implementation 24bpp w/ alpha */
inline static void gfx_vline_alpha_24(SDL_Surface *s, int x, int y, int y2, Uint8 alpha, Uint32 color)
{
	Uint8 *p = ((Uint8 *)s->pixels) + s->pitch * y + x*3;
	Uint8 *p2 = ((Uint8 *)s->pixels) + s->pitch * y2 + x*3;
	Uint32 c[3];
	Uint32 malpha = alpha;
	Uint32 ialpha = 255 - malpha;

	if(SDL_BYTEORDER == SDL_BIG_ENDIAN) {
		c[0] = ((color >> 16) & 0xff) * alpha;
		c[1] = ((color >> 8) & 0xff) * alpha;
		c[2] = (color & 0xff) * alpha;
	} else {
		c[0] = (color & 0xff) * alpha;
		c[1] = ((color >> 8) & 0xff) * alpha;
		c[2] = ((color >> 16) & 0xff) * alpha;
	}

	while(p<=p2) {
		*p = (((Uint32) *p) * ialpha + c[0]) >> 8;
		*(p+1) = (((Uint32) *(p+1)) * ialpha + c[1]) >> 8;
		*(p+2) = (((Uint32) *(p+2)) * ialpha + c[2]) >> 8;

		p += s->pitch;
	}
}

/* vertical line implementation 32bpp */
inline static void gfx_vline_32(SDL_Surface *s, int x, int y, int y2, Uint32 color)
{
	Uint8 *p = ((Uint8 *)s->pixels) + s->pitch * y + x*4;
	Uint8 *p2 = ((Uint8 *)s->pixels) + s->pitch * y2 + x*4;
	while(p<=p2) {
		*((Uint32*)p) = color;
		p += s->pitch;
	}
}

/* vertical line implementation 32bpp w/ alpha */
inline static void gfx_vline_alpha_32(SDL_Surface *s, int x, int y, int y2, Uint8 alpha, Uint32 color)
{
	Uint8 *p = ((Uint8 *)s->pixels) + s->pitch * y + x*4;
	Uint8 *p2 = ((Uint8 *)s->pixels) + s->pitch * y2 + x*4;
#ifdef BGR
	color = SDL_Swap32(color);
#endif
	Uint32 cg = color & 0x00ff00;
	Uint32 cng = color & 0xff00ff;
	Uint32 t1, t2;
	Uint32 ip;
	while(p<=p2) {
#ifdef BGR
		ip = SDL_Swap32(*((Uint32*) p));
#else
		ip = *((Uint32*) p);
#endif
		t2 = ip & 0xff00ff;
		t1 = cng - t2;
		t1 = (((t1*alpha) >> 8) + t2) & 0xff00ff;
		ip &= 0xff00;
		t2 = ((((cg - ip) * alpha) >> 8) + ip) & 0xff00;
#ifdef BGR
		*((Uint32*) p) = SDL_Swap32(t1 | t2);
#else
		*((Uint32*) p) = t1 | t2;
#endif
		p += s->pitch;
	}
}

/* vertical line */
void gfx_vline(SDL_Surface *s, int x, int y, int y2, Uint32 color)
{
	/* swap y and y2 as required */
	if(y > y2) {
		int t = y;
		y = y2;
		y2 = t;
	}

	/* clipping */

	int cx2 = s->clip_rect.x + s->clip_rect.w;
	int cy2 = s->clip_rect.y + s->clip_rect.h;

	if(y2 < s->clip_rect.y || y >= cy2) return;
	if(y < s->clip_rect.y) y = s->clip_rect.y;
	if(y2 >= cy2) y2 = cy2-1;
	if(x < s->clip_rect.x || x >= cx2) return;

	switch(s->format->BitsPerPixel) {
	case 8:
		gfx_vline_8(s, x, y, y2, color);
		break;
	case 15:
	case 16:
		gfx_vline_16(s, x, y, y2, color);
		break;
	case 24:
		gfx_vline_24(s, x, y, y2, color);
		break;
	case 32:
		gfx_vline_32(s, x, y, y2, color);
		break;
	default:
		abort();
	}
}

/* vertical line with alpha */
void gfx_vline_alpha(SDL_Surface *s, int x, int y, int y2, Uint8 alpha, Uint32 color)
{
	/* swap y and y2 as required */
	if(y > y2) {
		int t = y;
		y = y2;
		y2 = t;
	}

	/* clipping */

	int cx2 = s->clip_rect.x + s->clip_rect.w;
	int cy2 = s->clip_rect.y + s->clip_rect.h;

	if(y2 < s->clip_rect.y || y >= cy2) return;
	if(y < s->clip_rect.y) y = s->clip_rect.y;
	if(y2 >= cy2) y2 = cy2-1;
	if(x < s->clip_rect.x || x >= cx2) return;

	switch(s->format->BitsPerPixel) {
	case 8:
		gfx_vline_alpha_8(s, x, y, y2, alpha, color);
		break;
	case 15:
		gfx_vline_alpha_15(s, x, y, y2, alpha, color);
		break;
	case 16:
		gfx_vline_alpha_16(s, x, y, y2, alpha, color);
		break;
	case 24:
		gfx_vline_alpha_24(s, x, y, y2, alpha, color);
		break;
	case 32:
		gfx_vline_alpha_32(s, x, y, y2, alpha, color);
		break;
	default:
		abort();
	}
}

/* dot */
void gfx_dot(SDL_Surface *s, int x, int y, Uint32 color)
{
	int x2 = s->clip_rect.x + s->clip_rect.w;
	int y2 = s->clip_rect.y + s->clip_rect.h;
	if(x < s->clip_rect.x) return;
	if(y < s->clip_rect.y) return;
	if(x >= x2) return;
	if(y >= y2) return;

	Uint8 *p = (Uint8 *)s->pixels + y*s->pitch + x*s->format->BytesPerPixel;
	switch(s->format->BytesPerPixel) {
		case 1:
			*p = color;
			break;

		case 2:
			*(Uint16 *)p = color;
			break;

		case 3:
			if(SDL_BYTEORDER == SDL_BIG_ENDIAN) {
				p[0] = (color >> 16) & 0xff;
				p[1] = (color >> 8) & 0xff;
				p[2] = color & 0xff;
			} else {
				p[0] = color & 0xff;
				p[1] = (color >> 8) & 0xff;
				p[2] = (color >> 16) & 0xff;
			}
			break;

		case 4:
			*(Uint32 *)p = color;
			break;
	}
}

/* dot alpha */
void gfx_dot_alpha(SDL_Surface *s, int x, int y, Uint8 alpha, Uint32 color)
{
	int x2 = s->clip_rect.x + s->clip_rect.w;
	int y2 = s->clip_rect.y + s->clip_rect.h;
	if(x < s->clip_rect.x) return;
	if(y < s->clip_rect.y) return;
	if(x >= x2) return;
	if(y >= y2) return;

	Uint8 *p = (Uint8 *)s->pixels + y*s->pitch + x*s->format->BytesPerPixel;

	Uint32 ma, cg, cng, t1, t2;

	switch(s->format->BitsPerPixel) {
		case 8:
			ma = (alpha + 3) >> 5;
			cg = (color & 0x1c) >> 2;
			cng = color & 0xe3;
			t1 = *p & 0xe3;
			t2 = *p & 0x1c;
			*p = (((((cg - (t2 >> 2)) * ma) >> 1) + t2) & 0x1c) | (((((cng - t1) * ma) >> 3) + t1) & 0xe3);
			break;

		case 15:
			ma = (alpha + 7) >> 3;
			cg = (color & 0x03e0) >> 5;
			cng = color & 0x7c1f;
			t1 = *((Uint16*)p) & 0x7c1f;
			t2 = *((Uint16*)p) & 0x03e0;
			*((Uint16*)p) = ((((cg - (t2 >> 5)) * ma) + t2) & 0x03e0) | (((((cng - t1) * ma) >> 5) + t1) & 0x7c1f);
			break;

		case 16:
			ma = (alpha + 7) >> 3;
			cg = (color & 0x07e0) >> 5;
			cng = color & 0xf81f;
			t1 = *((Uint16*)p) & 0xf81f;
			t2 = *((Uint16*)p) & 0x07e0;
			*((Uint16*)p) = ((((cg - (t2 >> 5)) * ma) + t2) & 0x07e0) | (((((cng - t1) * ma) >> 5) + t1) & 0xf81f);
			break;

		case 24:
			ma = 255 - alpha;
			if(SDL_BYTEORDER == SDL_BIG_ENDIAN) {
				p[0] = (((Uint32) p[0]) * ma + ((color >> 16) & 0xff) * alpha) >> 8;
				p[1] = (((Uint32) p[1]) * ma + ((color >> 8) & 0xff) * alpha) >> 8;
				p[2] = (((Uint32) p[2]) * ma + (color & 0xff) * alpha) >> 8;
			} else {
				p[0] = (((Uint32) p[0]) * ma + (color & 0xff) * alpha) >> 8;
				p[1] = (((Uint32) p[1]) * ma + ((color >> 8) & 0xff) * alpha) >> 8;
				p[2] = (((Uint32) p[2]) * ma + ((color >> 16) & 0xff) * alpha) >> 8;
			}
			break;

		case 32:
#ifdef BGR
			color = SDL_Swap32(color);
#endif
			cg = color & 0x00ff00;
			cng = color & 0xff00ff;
#ifdef BGR
			ma = SDL_Swap32(*((Uint32*) p));
#else
			ma = *((Uint32*) p);
#endif
			t2 = ma & 0xff00ff;
			t1 = cng - t2;
			t1 = (((t1*alpha) >> 8) + t2) & 0xff00ff;
			ma &= 0xff00;
			t2 = ((((cg - ma) * alpha) >> 8) + ma) & 0xff00;
#ifdef BGR
			*((Uint32*) p) = SDL_Swap32(t1 | t2);
#else
			*((Uint32*) p) = t1 | t2;
#endif
			break;
	}
}

/* horizontal line */
void gfx_hline(SDL_Surface *s, int x1, int x2, int y, Uint32 color)
{
	/* swap x1 and x2 as required */
	if(x1 > x2) {
		int t = x1;
		x1 = x2;
		x2 = t;
	}

	/* clipping */

	int cx2 = s->clip_rect.x + s->clip_rect.w;
	int cy2 = s->clip_rect.y + s->clip_rect.h;

	if(x2 < s->clip_rect.x || x1 >= cx2) return;
	if(x1 < s->clip_rect.x) x1 = s->clip_rect.x;
	if(x2 >= cx2) x2 = cx2 - 1;
	if(y < s->clip_rect.y || y >= cy2) return;

	switch(s->format->BitsPerPixel) {
	case 8:
		gfx_hline_8(s, x1, x2, y, color);
		break;
	case 15:
	case 16:
		gfx_hline_16(s, x1, x2, y, color);
		break;
	case 24:
		gfx_hline_24(s, x1, x2, y, color);
		break;
	case 32:
		gfx_hline_32(s, x1, x2, y, color);
		break;
	default:
		abort();
	}
}

/* horizontal line with alpha  */
void gfx_hline_alpha(SDL_Surface *s, int x1, int x2, int y, Uint8 alpha, Uint32 color)
{
	/* swap x1 and x2 as required */
	if(x1 > x2) {
		int t = x1;
		x1 = x2;
		x2 = t;
	}

	/* clipping */

	int cx2 = s->clip_rect.x + s->clip_rect.w;
	int cy2 = s->clip_rect.y + s->clip_rect.h;

	if(x2 < s->clip_rect.x || x1 >= cx2) return;
	if(x1 < s->clip_rect.x) x1 = s->clip_rect.x;
	if(x2 >= cx2) x2 = cx2 - 1;
	if(y < s->clip_rect.y || y >= cy2) return;

	switch(s->format->BitsPerPixel) {
	case 8:
		gfx_hline_alpha_8(s, x1, x2, y, alpha, color);
		break;
	case 15:
		gfx_hline_alpha_15(s, x1, x2, y, alpha, color);
		break;
	case 16:
		gfx_hline_alpha_16(s, x1, x2, y, alpha, color);
		break;
	case 24:
		gfx_hline_alpha_24(s, x1, x2, y, alpha, color);
		break;
	case 32:
		gfx_hline_alpha_32(s, x1, x2, y, alpha, color);
		break;
	default:
		abort();
	}
}

/* rectangle */
void gfx_rect(SDL_Surface *s, int x1, int y1, int x2, int y2, Uint32 color)
{
	/* swap y1 and y2 as required.. */
	if(y1 > y2) {
		int t = y1;
		y1 = y2;
		y2 = t;
	}

	gfx_hline(s, x1, x2, y1, color);
	gfx_hline(s, x1, x2, y2, color);
	gfx_vline(s, x1, y1+1, y2-1, color);
	gfx_vline(s, x2, y1+1, y2-1, color);
}

/* box */
inline void gfx_box(SDL_Surface *s, int x1, int y1, int x2, int y2, Uint32 color)
{
	/* swap x1 and x2 as required */
	if(x1 > x2) {
		int t = x1;
		x1 = x2;
		x2 = t;
	}
	/* swap y1 and y2 as required */
	if(y1 > y2) {
		int t = y1;
		y1 = y2;
		y2 = t;
	}
	SDL_Rect dr = {x1, y1, (x2-x1+1), (y2-y1+1)};
	SDL_FillRect(s, &dr, color);
}

/* line drawing stuff adapted from SDL gfx */

/* --------- Clipping routines for line */

/* Clipping based heavily on code from                       */
/* http://www.ncsa.uiuc.edu/Vis/Graphics/src/clipCohSuth.c   */

#define CLIP_LEFT_EDGE   0x1
#define CLIP_RIGHT_EDGE  0x2
#define CLIP_BOTTOM_EDGE 0x4
#define CLIP_TOP_EDGE    0x8
#define CLIP_INSIDE(a)   (!a)
#define CLIP_REJECT(a,b) (a&b)
#define CLIP_ACCEPT(a,b) (!(a|b))

static int clipEncode(Sint16 x, Sint16 y, Sint16 left, Sint16 top, Sint16 right, Sint16 bottom)
{
    int code = 0;

    if (x < left) {
	code |= CLIP_LEFT_EDGE;
    } else if (x > right) {
	code |= CLIP_RIGHT_EDGE;
    }
    if (y < top) {
	code |= CLIP_TOP_EDGE;
    } else if (y > bottom) {
	code |= CLIP_BOTTOM_EDGE;
    }
    return code;
}

static int clipLine(SDL_Surface * dst, Sint16 * x1, Sint16 * y1, Sint16 * x2, Sint16 * y2)
{
    Sint16 left, right, top, bottom;
    int code1, code2;
    int draw = 0;
    Sint16 swaptmp;
    float m;

    /*
     * Get clipping boundary 
     */
    left = dst->clip_rect.x;
    right = dst->clip_rect.x + dst->clip_rect.w - 1;
    top = dst->clip_rect.y;
    bottom = dst->clip_rect.y + dst->clip_rect.h - 1;

    while (1) {
	code1 = clipEncode(*x1, *y1, left, top, right, bottom);
	code2 = clipEncode(*x2, *y2, left, top, right, bottom);
	if (CLIP_ACCEPT(code1, code2)) {
	    draw = 1;
	    break;
	} else if (CLIP_REJECT(code1, code2))
	    break;
	else {
	    if (CLIP_INSIDE(code1)) {
		swaptmp = *x2;
		*x2 = *x1;
		*x1 = swaptmp;
		swaptmp = *y2;
		*y2 = *y1;
		*y1 = swaptmp;
		swaptmp = code2;
		code2 = code1;
		code1 = swaptmp;
	    }
	    if (*x2 != *x1) {
		m = (*y2 - *y1) / (float) (*x2 - *x1);
	    } else {
		m = 1.0f;
	    }
	    if (code1 & CLIP_LEFT_EDGE) {
		*y1 += (Sint16) ((left - *x1) * m);
		*x1 = left;
	    } else if (code1 & CLIP_RIGHT_EDGE) {
		*y1 += (Sint16) ((right - *x1) * m);
		*x1 = right;
	    } else if (code1 & CLIP_BOTTOM_EDGE) {
		if (*x2 != *x1) {
		    *x1 += (Sint16) ((bottom - *y1) / m);
		}
		*y1 = bottom;
	    } else if (code1 & CLIP_TOP_EDGE) {
		if (*x2 != *x1) {
		    *x1 += (Sint16) ((top - *y1) / m);
		}
		*y1 = top;
	    }
	}
    }

    return draw;
}

/* ----- Line */

/* Non-alpha line drawing code adapted from routine          */
/* by Pete Shinners, pete@shinners.org                       */
/* Originally from pygame, http://pygame.seul.org            */

#define ABS(a) (((a)<0) ? -(a) : (a))

void gfx_line(SDL_Surface * dst, Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Uint32 color)
{
	int pixx, pixy;
	int x, y;
	int dx, dy;
	int sx, sy;
	int swaptmp;
	Uint8 *pixel;
	Uint8 p24c[3];

	/*
	 * Clip line and test if we have to draw 
	 */
	if (!(clipLine(dst, &x1, &y1, &x2, &y2))) return;

	/*
	 * Test for special cases of straight lines
	 */
	if (x1 == x2) {
		return gfx_vline(dst, x1, y1, y2, color);
	}
	if (y1 == y2) {
		return gfx_hline(dst, x1, x2, y1, color);
	}

	/*
	 * Variable setup 
	 */
	dx = x2 - x1;
	dy = y2 - y1;
	sx = (dx >= 0) ? 1 : -1;
	sy = (dy >= 0) ? 1 : -1;

	dx = sx * dx + 1;
	dy = sy * dy + 1;
	pixx = dst->format->BytesPerPixel;
	pixy = dst->pitch;
	pixel = ((Uint8 *) dst->pixels) + pixx * (int) x1 + pixy * (int) y1;
	pixx *= sx;
	pixy *= sy;
	if (dx < dy) {
	    swaptmp = dx;
	    dx = dy;
	    dy = swaptmp;
	    swaptmp = pixx;
	    pixx = pixy;
	    pixy = swaptmp;
	}

	/*
	 * Draw 
	 */
	x = 0;
	y = 0;
	switch (dst->format->BytesPerPixel) {
	case 1:
	    for (; x < dx; x++, pixel += pixx) {
		*pixel = color;
		y += dy;
		if (y >= dx) {
		    y -= dx;
		    pixel += pixy;
		}
	    }
	    break;
	case 2:
	    for (; x < dx; x++, pixel += pixx) {
		*(Uint16 *) pixel = color;
		y += dy;
		if (y >= dx) {
		    y -= dx;
		    pixel += pixy;
		}
	    }
	    break;
	case 3:
		if (SDL_BYTEORDER == SDL_BIG_ENDIAN) {
			p24c[0] = (color >> 16) & 0xff;
			p24c[1] = (color >> 8) & 0xff;
			p24c[2] = color & 0xff;
		} else {
			p24c[0] = color & 0xff;
			p24c[1] = (color >> 8) & 0xff;
			p24c[2] = (color >> 16) & 0xff;
		}
		for (; x < dx; x++, pixel += pixx) {
			pixel[0] = p24c[0];
			pixel[1] = p24c[1];
			pixel[2] = p24c[2];
			y += dy;
			if (y >= dx) {
				y -= dx;
				pixel += pixy;
			}
		}
	    break;
	default:
		/* 32 bpp */
	    for (; x < dx; x++, pixel += pixx) {
		*(Uint32 *) pixel = color;
		y += dy;
		if (y >= dx) {
		    y -= dx;
		    pixel += pixy;
		}
	    }
	    break;
	}
}

/* dashed line */
void gfx_line_dash(SDL_Surface * dst, Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Uint32 color, unsigned int dash)
{
	int pixx, pixy;
	int x, y;
	int dx, dy;
	int sx, sy;
	int swaptmp;
	Uint8 *pixel;
	Uint8 p24c[3];

	/*
	 * Clip line and test if we have to draw 
	 */
	if (!(clipLine(dst, &x1, &y1, &x2, &y2))) return;

	/*
	 * Variable setup 
	 */
	dx = x2 - x1;
	dy = y2 - y1;
	sx = (dx >= 0) ? 1 : -1;
	sy = (dy >= 0) ? 1 : -1;

	dx = sx * dx + 1;
	dy = sy * dy + 1;
	pixx = dst->format->BytesPerPixel;
	pixy = dst->pitch;
	pixel = ((Uint8 *) dst->pixels) + pixx * (int) x1 + pixy * (int) y1;
	pixx *= sx;
	pixy *= sy;
	if (dx < dy) {
	    swaptmp = dx;
	    dx = dy;
	    dy = swaptmp;
	    swaptmp = pixx;
	    pixx = pixy;
	    pixy = swaptmp;
	}

	/*
	 * Draw 
	 */
	x = 0;
	y = 0;
	switch (dst->format->BytesPerPixel) {
	case 1:
		for (; x < dx; x++, pixel += pixx) {
			if(dash & dashptr) *pixel = color;
			dashptr_advance();
			y += dy;
			if (y >= dx) {
				y -= dx;
				pixel += pixy;
			}
		}
		break;
	case 2:
		for (; x < dx; x++, pixel += pixx) {
			if(dash & dashptr) *(Uint16 *) pixel = color;
			dashptr_advance();
			y += dy;
			if (y >= dx) {
				y -= dx;
				pixel += pixy;
			}
		}
		break;
	case 3:
		if (SDL_BYTEORDER == SDL_BIG_ENDIAN) {
			p24c[0] = (color >> 16) & 0xff;
			p24c[1] = (color >> 8) & 0xff;
			p24c[2] = color & 0xff;
		} else {
			p24c[0] = color & 0xff;
			p24c[1] = (color >> 8) & 0xff;
			p24c[2] = (color >> 16) & 0xff;
		}
		for (; x < dx; x++, pixel += pixx) {
			if(dash & dashptr) {
				pixel[0] = p24c[0];
				pixel[1] = p24c[1];
				pixel[2] = p24c[2];
			}
			dashptr_advance();
			y += dy;
			if (y >= dx) {
				y -= dx;
				pixel += pixy;
			}
		}
	    break;
	default:
		/* 32 bpp */
		for (; x < dx; x++, pixel += pixx) {
			if(dash & dashptr) *(Uint32 *) pixel = color;
			dashptr_advance();
			y += dy;
			if (y >= dx) {
				y -= dx;
				pixel += pixy;
			}
		}
		break;
	}
}

/* line with alpha */
void gfx_line_alpha(SDL_Surface * dst, Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Uint8 alpha, Uint32 color)
{
	int pixx, pixy;
	int x, y;
	int dx, dy;
	int sx, sy;
	int swaptmp;
	Uint8 *pixel;
	Uint32 p24c[3];

	/*
	 * Clip line and test if we have to draw 
	 */
	if (!(clipLine(dst, &x1, &y1, &x2, &y2))) return;

	/*
	 * Test for special cases of straight lines
	 */
	if (x1 == x2) {
		return gfx_vline_alpha(dst, x1, y1, y2, alpha, color);
	}
	if (y1 == y2) {
		return gfx_hline_alpha(dst, x1, x2, y1, alpha, color);
	}

	/*
	 * Variable setup 
	 */
	dx = x2 - x1;
	dy = y2 - y1;
	sx = (dx >= 0) ? 1 : -1;
	sy = (dy >= 0) ? 1 : -1;

	dx = sx * dx + 1;
	dy = sy * dy + 1;
	pixx = dst->format->BytesPerPixel;
	pixy = dst->pitch;
	pixel = ((Uint8 *) dst->pixels) + pixx * (int) x1 + pixy * (int) y1;
	pixx *= sx;
	pixy *= sy;
	if (dx < dy) {
	    swaptmp = dx;
	    dx = dy;
	    dy = swaptmp;
	    swaptmp = pixx;
	    pixx = pixy;
	    pixy = swaptmp;
	}

	/*
	 * Draw 
	 */
	x = 0;
	y = 0;

	Uint32 ma;
	Uint32 cg, cng;
	Uint32 t1, t2;

	switch (dst->format->BitsPerPixel) {
	case 8:
		ma = (alpha + 3) >> 5;
		cg = (color & 0x1c) >> 2;
		cng = color & 0xe3;
		for (; x < dx; x++, pixel += pixx) {
			t1 = *pixel & 0xe3;
			t2 = *pixel & 0x1c;
			*pixel = (((((cg - (t2 >> 2)) * ma) >> 1) + t2) & 0x1c) | (((((cng - t1) * ma) >> 3) + t1) & 0xe3);
			y += dy;
			if (y >= dx) {
				y -= dx;
				pixel += pixy;
			}
		}
		break;
	case 15:
		ma = (alpha + 7) >> 3;
		cg = (color & 0x03e0) >> 5;
		cng = color & 0x7c1f;
		for (; x < dx; x++, pixel += pixx) {
			t1 = *((Uint16*)pixel) & 0x7c1f;
			t2 = *((Uint16*)pixel) & 0x03e0;
			*((Uint16*)pixel) = ((((cg - (t2 >> 5)) * ma) + t2) & 0x03e0) | (((((cng - t1) * ma) >> 5) + t1) & 0x7c1f);
			y += dy;
			if (y >= dx) {
				y -= dx;
				pixel += pixy;
			}
		}
		break;
	case 16:
		ma = (alpha + 7) >> 3;
		cg = (color & 0x07e0) >> 5;
		cng = color & 0xf81f;
		for (; x < dx; x++, pixel += pixx) {
			t1 = *((Uint16*)pixel) & 0xf81f;
			t2 = *((Uint16*)pixel) & 0x07e0;
			*((Uint16*)pixel) = ((((cg - (t2 >> 5)) * ma) + t2) & 0x07e0) | (((((cng - t1) * ma) >> 5) + t1) & 0xf81f);
			y += dy;
			if (y >= dx) {
				y -= dx;
				pixel += pixy;
			}
		}
		break;
	case 24:
		ma = 255 - alpha;
		if (SDL_BYTEORDER == SDL_BIG_ENDIAN) {
			p24c[0] = ((color >> 16) & 0xff) * alpha;
			p24c[1] = ((color >> 8) & 0xff) * alpha;
			p24c[2] = (color & 0xff) * alpha;
		} else {
			p24c[0] = (color & 0xff) * alpha;
			p24c[1] = ((color >> 8) & 0xff) * alpha;
			p24c[2] = ((color >> 16) & 0xff) * alpha;
		}
		for (; x < dx; x++, pixel += pixx) {
			pixel[0] = (pixel[0] * ma + p24c[0]) >> 8;
			pixel[1] = (pixel[1] * ma + p24c[1]) >> 8;
			pixel[2] = (pixel[2] * ma + p24c[2]) >> 8;
			y += dy;
			if (y >= dx) {
				y -= dx;
				pixel += pixy;
			}
		}
		break;
	default:
		/* 32 bpp */
#ifdef BGR
		color = SDL_Swap32(color);
#endif
		cg = color & 0x00ff00;
		cng = color & 0xff00ff;
		for (; x < dx; x++, pixel += pixx) {
#ifdef BGR
			ma = SDL_Swap32(*((Uint32*) pixel));
#else
			ma = *((Uint32*) pixel);
#endif
			t2 = ma & 0xff00ff;
			t1 = cng - t2;
			t1 = (((t1*alpha) >> 8) + t2) & 0xff00ff;
			ma &= 0xff00;
			t2 = ((((cg - ma) * alpha) >> 8) + ma) & 0xff00;
#ifdef BGR
			*((Uint32*) pixel) = SDL_Swap32(t1 | t2);
#else
			*((Uint32*) pixel) = t1 | t2;
#endif
			y += dy;
			if (y >= dx) {
				y -= dx;
				pixel += pixy;
			}
		}
		break;
	}
}

/* dashed line with alpha */
void gfx_line_alpha_dash(SDL_Surface * dst, Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Uint8 alpha, Uint32 color, unsigned int dash)
{
	int pixx, pixy;
	int x, y;
	int dx, dy;
	int sx, sy;
	int swaptmp;
	Uint8 *pixel;
	Uint32 p24c[3];

	/*
	 * Clip line and test if we have to draw 
	 */
	if (!(clipLine(dst, &x1, &y1, &x2, &y2))) return;

	/*
	 * Variable setup 
	 */
	dx = x2 - x1;
	dy = y2 - y1;
	sx = (dx >= 0) ? 1 : -1;
	sy = (dy >= 0) ? 1 : -1;

	dx = sx * dx + 1;
	dy = sy * dy + 1;
	pixx = dst->format->BytesPerPixel;
	pixy = dst->pitch;
	pixel = ((Uint8 *) dst->pixels) + pixx * (int) x1 + pixy * (int) y1;
	pixx *= sx;
	pixy *= sy;
	if (dx < dy) {
	    swaptmp = dx;
	    dx = dy;
	    dy = swaptmp;
	    swaptmp = pixx;
	    pixx = pixy;
	    pixy = swaptmp;
	}

	/*
	 * Draw 
	 */
	x = 0;
	y = 0;

	Uint32 ma;
	Uint32 cg, cng;
	Uint32 t1, t2;

	switch (dst->format->BitsPerPixel) {
	case 8:
		ma = (alpha + 3) >> 5;
		cg = (color & 0x1c) >> 2;
		cng = color & 0xe3;
		for (; x < dx; x++, pixel += pixx) {
			if(dash & dashptr) {
				t1 = *pixel & 0xe3;
				t2 = *pixel & 0x1c;
				*pixel = (((((cg - (t2 >> 2)) * ma) >> 1) + t2) & 0x1c) | (((((cng - t1) * ma) >> 3) + t1) & 0xe3);
			}
			dashptr_advance();
			y += dy;
			if (y >= dx) {
				y -= dx;
				pixel += pixy;
			}
		}
		break;
	case 15:
		ma = (alpha + 7) >> 3;
		cg = (color & 0x03e0) >> 5;
		cng = color & 0x7c1f;
		for (; x < dx; x++, pixel += pixx) {
			if(dash & dashptr) {
				t1 = *((Uint16*)pixel) & 0x7c1f;
				t2 = *((Uint16*)pixel) & 0x03e0;
				*((Uint16*)pixel) = ((((cg - (t2 >> 5)) * ma) + t2) & 0x03e0) | (((((cng - t1) * ma) >> 5) + t1) & 0x7c1f);
			}
			dashptr_advance();
			y += dy;
			if (y >= dx) {
				y -= dx;
				pixel += pixy;
			}
		}
		break;
	case 16:
		ma = (alpha + 7) >> 3;
		cg = (color & 0x07e0) >> 5;
		cng = color & 0xf81f;
		for (; x < dx; x++, pixel += pixx) {
			if(dash & dashptr) {
				t1 = *((Uint16*)pixel) & 0xf81f;
				t2 = *((Uint16*)pixel) & 0x07e0;
				*((Uint16*)pixel) = ((((cg - (t2 >> 5)) * ma) + t2) & 0x07e0) | (((((cng - t1) * ma) >> 5) + t1) & 0xf81f);
			}
			dashptr_advance();
			y += dy;
			if (y >= dx) {
				y -= dx;
				pixel += pixy;
			}
		}
		break;
	case 24:
		ma = 255 - alpha;
		if (SDL_BYTEORDER == SDL_BIG_ENDIAN) {
			p24c[0] = ((color >> 16) & 0xff) * alpha;
			p24c[1] = ((color >> 8) & 0xff) * alpha;
			p24c[2] = (color & 0xff) * alpha;
		} else {
			p24c[0] = (color & 0xff) * alpha;
			p24c[1] = ((color >> 8) & 0xff) * alpha;
			p24c[2] = ((color >> 16) & 0xff) * alpha;
		}
		for (; x < dx; x++, pixel += pixx) {
			if(dash & dashptr) {
				pixel[0] = (pixel[0] * ma + p24c[0]) >> 8;
				pixel[1] = (pixel[1] * ma + p24c[1]) >> 8;
				pixel[2] = (pixel[2] * ma + p24c[2]) >> 8;
			}
			dashptr_advance();
			y += dy;
			if (y >= dx) {
				y -= dx;
				pixel += pixy;
			}
		}
		break;
	default:
		/* 32 bpp */
#ifdef BGR
		color = SDL_Swap32(color);
#endif
		cg = color & 0x00ff00;
		cng = color & 0xff00ff;
		for (; x < dx; x++, pixel += pixx) {
			if(dash & dashptr) {
#ifdef BGR
				ma = SDL_Swap32(*((Uint32*) pixel));
#else
				ma = *((Uint32*) pixel);
#endif
				t2 = ma & 0xff00ff;
				t1 = cng - t2;
				t1 = (((t1*alpha) >> 8) + t2) & 0xff00ff;
				ma &= 0xff00;
				t2 = ((((cg - ma) * alpha) >> 8) + ma) & 0xff00;
#ifdef BGR
				*((Uint32*) pixel) = SDL_Swap32(t1 | t2);
#else
				*((Uint32*) pixel) = t1 | t2;
#endif
			}
			dashptr_advance();
			y += dy;
			if (y >= dx) {
				y -= dx;
				pixel += pixy;
			}
		}
		break;
	}
}

#include "font.h"

/* size of a single character in the compiled in font */
int font_w, font_h;

/* number of bytes in a font row */
int font_pitch;


#define CLIP_STACK_DEPTH 16
static SDL_Rect clip_stack[CLIP_STACK_DEPTH];
static int clip_stack_ptr;

#define MIN(x,y) ((x)<(y) ? (x):(y))
#define MAX(x,y) ((x)>(y) ? (x):(y))

void d_clip(int x, int y, int w, int h)
{
	assert(clip_stack_ptr < CLIP_STACK_DEPTH);

	if(clip_stack_ptr > 0) {
		SDL_Rect *r = &main_surface->clip_rect;
		int rx2 = r->x + r->w;
		int ry2 = r->y + r->h;
		int x2 = x + w;
		int y2 = y + h;
		x = MAX(x, r->x);
		y = MAX(y, r->y);
		int nx2 = MIN(x2, rx2);
		int ny2 = MIN(y2, ry2);
		w = nx2 - x;
		h = ny2 - y;
		if(w < 0) w = 0;
		if(h < 0) h = 0;
	}

	clip_stack[clip_stack_ptr ++] = main_surface->clip_rect;
	main_surface->clip_rect.x = x;
	main_surface->clip_rect.y = y;
	main_surface->clip_rect.w = w;
	main_surface->clip_rect.h = h;
}

void d_clip_debug()
{
	int i;
	SDL_Rect *r = clip_stack;
	for(i=0; i<clip_stack_ptr; i++, r++)  {
		printf("%2d %4d,%4d %4d,%4d\n", i, r->x, r->y, r->w, r->h);
	}
	r = &main_surface->clip_rect;
	printf("CR %4d,%4d %4d,%4d\n", r->x, r->y, r->w, r->h);
}

void d_unclip()
{
	assert(clip_stack_ptr > 0);
	main_surface->clip_rect = clip_stack[--clip_stack_ptr];
}

void d_clear(Uint32 color)
{
	SDL_FillRect(main_surface, NULL, color);
}

void d_begin()
{
	SDL_LockSurface(main_surface);
}

void d_end()
{
	SDL_UnlockSurface(main_surface);
	SDL_Flip(main_surface);
}

Uint32 mapcolori(int r, int g, int b)
{
	return SDL_MapRGB(main_surface->format, r, g, b);
}

Uint32 mapcolorf(float r, float g, float b)
{
	inline int cb(float x) {
		int z = x * 255.0;
		if(z < 0) z = 0;
		else if(z > 255) z = 255;
		return z;
	}

	return mapcolori(cb(r), cb(g), cb(b));
}

/* initialize gfx */
void gfx_init()
{
	font_w = font_width / 16;
	font_h = font_height / 16;
	font_pitch = font_width / 8;
	if(font_width%8) font_pitch ++;
}

/* use surface for drawing, use software routines */
void gfx_surface_sw(SDL_Surface *s)
{
	main_surface = s;
}

/* draw a scanline of text */
static void gfx_textln(SDL_Surface *s, Uint8 *p, int x, const char *str, const char *end, int cx, int cy, Uint32 color)
{
	unsigned char c = *str;
	if(!c || str == end) return;

	int clipright = s->clip_rect.x + s->clip_rect.w;

	/* calculate current position in font bitmap */
	int fy = (c>>4) * font_h + cy;
	int fx = (c&15) * font_w + cx;

	/* current byte in font bitmap */
	unsigned char *fp = font_bits + fy * font_pitch + (fx>>3);
	/* mask to access current pixel */
	unsigned char mask = 1 << (fx & 7);

	/* pixel advance logic
	 * returns 1 if scanline is over. */
	inline int advance() {
		x++;
		if(x == clipright) return 1;

		mask = mask << 1;
		if(!mask) {
			/* finished rendering current byte in font bitmap */
			mask = 1;
			fp++;
		}

		cx++;
		if(cx == font_w) {
			/* finished rendering current character */

			str ++;
			if(str == end) return 1;
			c = *str;
			if(!c) return 1;

			cx = 0;

			/* calculate current position in font bitmap */
			fy = (c>>4) * font_h + cy;
			fx = (c&15) * font_w;

			/* current byte in font bitmap */
			fp = font_bits + fy * font_pitch + (fx>>3);
			/* mask to access current pixel */
			mask = 1 << (fx & 7);
		}

		return 0;
	}

	switch(s->format->BytesPerPixel) {
	case 1:
		for(;;) {
			if(*fp & mask) *p = color;
			p++;
			if(advance()) return;
		}
	case 2:
		for(;;) {
			if(*fp & mask) *((Uint16*)p) = color;
			p+=2;
			if(advance()) return;
		}
	case 3:
		for(;;) {
			if(*fp & mask) {
				if (SDL_BYTEORDER == SDL_BIG_ENDIAN) {
					*p++ = (color >> 16) & 0xff;
					*p++ = (color >> 8) & 0xff;
					*p++ = color & 0xff;
				} else {
					*p++ = color & 0xff;
					*p++ = (color >> 8) & 0xff;
					*p++ = (color >> 16) & 0xff;
				}
			} else p+=3;

			if(advance()) return;
		}
	case 4:
		for(;;) {
			if(*fp & mask) *((Uint32*)p) = color;
			p += 4;
			if(advance()) return;
		}
	default:
		abort();
	}
}

/* draw a scanline of text with alpha */
static void gfx_textln_alpha(SDL_Surface *s, Uint8 *p, int x, const char *str, const char *end, int cx, int cy, Uint8 alpha, Uint32 color)
{
	unsigned char c = *str;
	if(!c || str == end) return;

	int clipright = s->clip_rect.x + s->clip_rect.w;

	/* calculate current position in font bitmap */
	int fy = (c>>4) * font_h + cy;
	int fx = (c&15) * font_w + cx;

	/* current byte in font bitmap */
	unsigned char *fp = font_bits + fy * font_pitch + (fx>>3);
	/* mask to access current pixel */
	unsigned char mask = 1 << (fx & 7);

	/* pixel advance logic
	 * returns 1 if scanline is over. */
	inline int advance() {
		x++;
		if(x == clipright) return 1;

		mask = mask << 1;
		if(!mask) {
			/* finished rendering current byte in font bitmap */
			mask = 1;
			fp++;
		}

		cx++;
		if(cx == font_w) {
			/* finished rendering current character */

			str ++;
			if(str == end) return 1;
			c = *str;
			if(!c) return 1;

			cx = 0;

			/* calculate current position in font bitmap */
			fy = (c>>4) * font_h + cy;
			fx = (c&15) * font_w;

			/* current byte in font bitmap */
			fp = font_bits + fy * font_pitch + (fx>>3);
			/* mask to access current pixel */
			mask = 1 << (fx & 7);
		}

		return 0;
	}

	Uint32 ma, cg, cng, t1, t2;
	Uint32 p24c[3];

	switch(s->format->BitsPerPixel) {
	case 8:
		ma = (alpha + 3) >> 5;
		cg = (color & 0x1c) >> 2;
		cng = color & 0xe3;
		for(;;) {
			if(*fp & mask) {
				t1 = *p & 0xe3;
				t2 = *p & 0x1c;
				*p = (((((cg - (t2 >> 2)) * ma) >> 1) + t2) & 0x1c) | (((((cng - t1) * ma) >> 3) + t1) & 0xe3);
			}
			p++;
			if(advance()) return;
		}
	case 15:
		ma = (alpha + 7) >> 3;
		cg = (color & 0x03e0) >> 5;
		cng = color & 0x7c1f;
		for(;;) {
			if(*fp & mask) {
				t1 = *((Uint16*)p) & 0x7c1f;
				t2 = *((Uint16*)p) & 0x03e0;
				*((Uint16*)p) = ((((cg - (t2 >> 5)) * ma) + t2) & 0x03e0) | (((((cng - t1) * ma) >> 5) + t1) & 0x7c1f);
			}
			p+=2;
			if(advance()) return;
		}
	case 16:
		ma = (alpha + 7) >> 3;
		cg = (color & 0x07e0) >> 5;
		cng = color & 0xf81f;
		for(;;) {
			if(*fp & mask) {
				t1 = *((Uint16*)p) & 0xf81f;
				t2 = *((Uint16*)p) & 0x07e0;
				*((Uint16*)p) = ((((cg - (t2 >> 5)) * ma) + t2) & 0x07e0) | (((((cng - t1) * ma) >> 5) + t1) & 0xf81f);
			}
			p+=2;
			if(advance()) return;
		}
	case 24:
		ma = 255 - alpha;
		if (SDL_BYTEORDER == SDL_BIG_ENDIAN) {
			p24c[0] = ((color >> 16) & 0xff) * alpha;
			p24c[1] = ((color >> 8) & 0xff) * alpha;
			p24c[2] = (color & 0xff) * alpha;
		} else {
			p24c[0] = (color & 0xff) * alpha;
			p24c[1] = ((color >> 8) & 0xff) * alpha;
			p24c[2] = ((color >> 16) & 0xff) * alpha;
		}

		for(;;) {
			if(*fp & mask) {
				*p = (((Uint32) *p) * ma + p24c[0]) >> 8; p++;
				*p = (((Uint32) *p) * ma + p24c[1]) >> 8; p++;
				*p = (((Uint32) *p) * ma + p24c[2]) >> 8; p++;
			} else p+=3;

			if(advance()) return;
		}
	case 32:
#ifdef BGR
		color = SDL_Swap32(color);
#endif
		cg = color & 0x00ff00;
		cng = color & 0xff00ff;
		for(;;) {
			if(*fp & mask) {
#ifdef BGR
				ma = SDL_Swap32(*((Uint32*) p));
#else
				ma = *((Uint32*) p);
#endif
				t2 = ma & 0xff00ff;
				t1 = cng - t2;
				t1 = (((t1*alpha) >> 8) + t2) & 0xff00ff;
				ma &= 0xff00;
				t2 = ((((cg - ma) * alpha) >> 8) + ma) & 0xff00;
#ifdef BGR
				*((Uint32*) p) = SDL_Swap32(t1 | t2);
#else
				*((Uint32*) p) = t1 | t2;
#endif
			}
			p += 4;
			if(advance()) return;
		}
	default:
		abort();
	}
}

/* draw a string */
void gfx_text(SDL_Surface *s, int x, int y, const char *str, const char *end, Uint8 alpha, Uint32 color)
{
	if(!*str) return;

	int cx, cy; /* character cell position */

	int clx2 = s->clip_rect.x + s->clip_rect.w;
	int cly2 = s->clip_rect.y + s->clip_rect.h;

	/* do some clipping */
	if(x >= clx2 || y >= cly2) return;

	if(x < s->clip_rect.x) {
		cx = (s->clip_rect.x-x) % font_w;
		int sk = (s->clip_rect.x-x) / font_w;
		while(sk>0) {
			str++;
			if(str == end || !*str) return;
			sk--;
		}
		x = s->clip_rect.x;
	} else cx = 0;

	if(y < s->clip_rect.y) {
		if(y < s->clip_rect.y-font_h+1) return;
		cy = (s->clip_rect.y-y);
		y = s->clip_rect.y;
	} else cy = 0;

	/* pixel offset */
	Uint8 *p = ((Uint8*)s->pixels) + y * s->pitch + x * s->format->BytesPerPixel;

	if(alpha == 255) {
		while(cy < font_h) {
			gfx_textln(s, p, x, str, end, cx, cy, color);
			y ++;
			if(y == cly2) return;
			cy ++;
			p += s->pitch;
		}
	} else {
		while(cy < font_h) {
			gfx_textln_alpha(s, p, x, str, end, cx, cy, alpha, color);
			y ++;
			if(y == cly2) return;
			cy ++;
			p += s->pitch;
		}
	}
}

/* draw multi-line string */
void d_ml_text(int x, int y, const char *str, Uint8 alpha, Uint32 color)
{
	const char *p;
	for(p = str; p;) {
		char *end = strchr(p, '\n');
		gfx_text(main_surface, x, y, p, end, alpha, color);
		p = end;
		if(p) {
			p ++;
			y += font_h;
		}
	}
}

/* friendly clipping that watches for surface boundaries */
void gfx_clip(int x, int y, int w, int h)
{
	if(w < 0) w = 0;
	if(h < 0) h = 0;
	if(x < 0) { w -= x; x = 0; }
	if(y < 0) { h -= y; y = 0; }
	if(x + w >= main_surface->w)
		w = main_surface->w - x;
	if(y + h >= main_surface->h)
		h = main_surface->h - y;
	d_clip(x, y, w, h);
}

void d_line_thick(Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, float thick, Uint8 alpha, Uint32 color)
{
	int xw = abs(x2-x1);
	int yw = abs(y2-y1);
	int px = (int) (thick+0.5);
	int bf = (int) (thick/2.0);
	int i;

	if(xw < yw) {
		x1 -= bf;
		x2 -= bf;
		if(alpha == 255) {
			for(i=0; i<px; i++) {
				gfx_line(main_surface, x1, y1, x2, y2, color);
				x1 ++;
				x2 ++;
			}
		} else {
			for(i=0; i<px; i++) {
				gfx_line_alpha(main_surface, x1, y1, x2, y2, alpha, color);
				x1 ++;
				x2 ++;
			}
		}
	} else {
		y1 -= bf;
		y2 -= bf;
		if(alpha == 255) {
			for(i=0; i<px; i++) {
				gfx_line(main_surface, x1, y1, x2, y2, color);
				y1 ++;
				y2 ++;
			}
		} else {
			for(i=0; i<px; i++) {
				gfx_line_alpha(main_surface, x1, y1, x2, y2, alpha, color);
				y1 ++;
				y2 ++;
			}
		}
	}
}

void d_line_thick_dash(Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, float thick, Uint8 alpha, Uint32 color, unsigned int dash)
{
	int xw = abs(x2-x1);
	int yw = abs(y2-y1);
	int px = (int) (thick+0.5);
	int bf = (int) (thick/2.0);
	int i;

	unsigned int dashptrs = dashptr;

	if(xw < yw) {
		x1 -= bf;
		x2 -= bf;
		if(alpha == 255) {
			for(i=0; i<px; i++) {
				dashptr = dashptrs;
				gfx_line_dash(main_surface, x1, y1, x2, y2, color, dash);
				x1 ++;
				x2 ++;
			}
		} else {
			for(i=0; i<px; i++) {
				dashptr = dashptrs;
				gfx_line_alpha_dash(main_surface, x1, y1, x2, y2, alpha, color, dash);
				x1 ++;
				x2 ++;
			}
		}
	} else {
		y1 -= bf;
		y2 -= bf;
		if(alpha == 255) {
			for(i=0; i<px; i++) {
				dashptr = dashptrs;
				gfx_line_dash(main_surface, x1, y1, x2, y2, color, dash);
				y1 ++;
				y2 ++;
			}
		} else {
			for(i=0; i<px; i++) {
				dashptr = dashptrs;
				gfx_line_alpha_dash(main_surface, x1, y1, x2, y2, alpha, color, dash);
				y1 ++;
				y2 ++;
			}
		}
	}
}

typedef struct {
	SDL_Surface *surf;
	Uint32 color;
	int alpha;
} fillpoly_cb_data_t;

static void fillpoly_h32(int y, poly_hline_segment_t *i, poly_hline_segment_t *end, void *data)
{
	if(i == end) return;

	Uint32 color;
	Uint32 color1 = ((fillpoly_cb_data_t *) data)->color;
	SDL_Surface *s = ((fillpoly_cb_data_t *) data)->surf;

#ifdef BGR
	color = SDL_Swap32(color1);
#else
	color = color1;
#endif

	Uint32 cg = color & 0x00ff00;
	Uint32 cng = color & 0xff00ff;

	Uint8 *p = ((Uint8 *)s->pixels) + s->pitch * y;

	for(; i < end; i++) {
		Uint8 *pi = p + i->x1 * 4;
		Uint8 *p2 = p + i->x2 * 4;

		if(i->alpha == 255) {
			while(pi < p2) {
				*((Uint32*) pi) = color1;
				pi += 4;
			}
		} else {
			while(pi < p2) {
				Uint32 ip, t1, t2;

#ifdef BGR
				ip = SDL_Swap32(*((Uint32*) pi));
#else
				ip = *((Uint32*) pi);
#endif

				t2 = ip & 0xff00ff;
				t1 = cng - t2;
				t1 = (((t1*i->alpha) >> 8) + t2) & 0xff00ff;
				ip &= 0xff00;
				t2 = ((((cg - ip) * i->alpha) >> 8) + ip) & 0xff00;
#ifdef BGR
				*((Uint32*) pi) = SDL_Swap32(t1 | t2);
#else
				*((Uint32*) pi) = t1 | t2;
#endif

				pi += 4;

			}
		}
	}
}

#define FILLPOLY_AA_CB(name, hline, hlinea) \
	static void name(int y, poly_hline_segment_t *i, poly_hline_segment_t *end, void *data) \
	{ \
		Uint32 color = ((fillpoly_cb_data_t *) data)->color; \
		SDL_Surface *surf = ((fillpoly_cb_data_t *) data)->surf; \
		for(; i < end; i++) { \
			/* printf("span y=%d x %d .. %d alpha %d\n", y, i->x1, i->x2, i->alpha); */ \
			if(i->alpha == 255) \
				hline(surf, i->x1, i->x2 - 1, y, color); \
			else \
				hlinea(surf, i->x1, i->x2 - 1, y, i->alpha, color); \
		} \
	}

#define FILLPOLY_AA_CB_A(name, hline, hlinea) \
	static void name(int y, poly_hline_segment_t *i, poly_hline_segment_t *end, void *data) \
	{ \
		Uint32 color = ((fillpoly_cb_data_t *) data)->color; \
		SDL_Surface *surf = ((fillpoly_cb_data_t *) data)->surf; \
		int alpha = ((fillpoly_cb_data_t *) data)->alpha; \
		for(; i < end; i++) { \
			uint8_t a = ((int) alpha * (int) i->alpha) / 255; \
			if(a == 255) \
				hline(surf, i->x1, i->x2 - 1, y, color); \
			else \
				hlinea(surf, i->x1, i->x2 - 1, y, a, color); \
		} \
	}

FILLPOLY_AA_CB(fillpoly_h8, gfx_hline_8, gfx_hline_alpha_8)
FILLPOLY_AA_CB(fillpoly_h15, gfx_hline_16, gfx_hline_alpha_15)
FILLPOLY_AA_CB(fillpoly_h16, gfx_hline_16, gfx_hline_alpha_16)
FILLPOLY_AA_CB(fillpoly_h24, gfx_hline_24, gfx_hline_alpha_24)
// FILLPOLY_AA_CB(fillpoly_h32, gfx_hline_32, gfx_hline_alpha_32)

FILLPOLY_AA_CB_A(fillpoly_ha8, gfx_hline_8, gfx_hline_alpha_8)
FILLPOLY_AA_CB_A(fillpoly_ha15, gfx_hline_16, gfx_hline_alpha_15)
FILLPOLY_AA_CB_A(fillpoly_ha16, gfx_hline_16, gfx_hline_alpha_16)
FILLPOLY_AA_CB_A(fillpoly_ha24, gfx_hline_24, gfx_hline_alpha_24)
FILLPOLY_AA_CB_A(fillpoly_ha32, gfx_hline_32, gfx_hline_alpha_32)

void gfx_fillpoly_aa(SDL_Surface *surf, int n,int pt[][2],Uint8 alpha,Uint32 color)
{
	poly_hline_a_t hline;

	if(!alpha) return;

	if(alpha == 255) {
		switch(surf->format->BitsPerPixel) {
		case 8: hline = fillpoly_h8; break;
		case 15: hline = fillpoly_h15; break;
		case 16: hline = fillpoly_h16; break;
		case 24: hline = fillpoly_h24; break;
		case 32: hline = fillpoly_h32; break;
		default:
			 abort();
		}
	} else {
		switch(surf->format->BitsPerPixel) {
		case 8: hline = fillpoly_ha8; break;
		case 15: hline = fillpoly_ha15; break;
		case 16: hline = fillpoly_ha16; break;
		case 24: hline = fillpoly_ha24; break;
		case 32: hline = fillpoly_ha32; break;
		default:
			 abort();
		}
	}

	fillpoly_cb_data_t data;
	data.surf = surf;
	data.color = color;
	data.alpha = alpha;

	scanpoly_a(surf->clip_rect.x, (surf->clip_rect.x + surf->clip_rect.w) - 1, surf->clip_rect.y, surf->clip_rect.y + surf->clip_rect.h - 1, n, pt, hline, &data);
}

void gfx_fillpoly(SDL_Surface *surf, int n,int pt[][2],Uint8 alpha,Uint32 color)
{
	poly_hline_t hline;

	void h8(int x1, int x2, int y) {
		gfx_hline_8(surf, x1, x2, y, color);
	}

	void h16(int x1, int x2, int y) {
		gfx_hline_16(surf, x1, x2, y, color);
	}

	void h24(int x1, int x2, int y) {
		gfx_hline_24(surf, x1, x2, y, color);
	}

	void h32(int x1, int x2, int y) {
		gfx_hline_32(surf, x1, x2, y, color);
	}

	void ha8(int x1, int x2, int y) {
		gfx_hline_alpha_8(surf, x1, x2, y, alpha, color);
	}

	void ha15(int x1, int x2, int y) {
		gfx_hline_alpha_15(surf, x1, x2, y, alpha, color);
	}

	void ha16(int x1, int x2, int y) {
		gfx_hline_alpha_16(surf, x1, x2, y, alpha, color);
	}

	void ha24(int x1, int x2, int y) {
		gfx_hline_alpha_24(surf, x1, x2, y, alpha, color);
	}

	void ha32(int x1, int x2, int y) {
		gfx_hline_alpha_32(surf, x1, x2, y, alpha, color);
	}

	if(!alpha) return;

	if(alpha == 255) {
		switch(surf->format->BitsPerPixel) {
		case 8: hline = h8; break;
		case 15:
		case 16:
				hline = h16; break;
		case 24: hline = h24; break;
		case 32: hline = h32; break;
		default:
			 abort();
		}
	} else {
		switch(surf->format->BitsPerPixel) {
		case 8: hline = ha8; break;
		case 15: hline = ha15; break;
		case 16: hline = ha16; break;
		case 24: hline = ha24; break;
		case 32: hline = ha32; break;
		default:
			 abort();
		}
	}

	scanpoly(surf->clip_rect.x, surf->clip_rect.x + surf->clip_rect.w - 1, surf->clip_rect.y, surf->clip_rect.y + surf->clip_rect.h - 1, n, pt, hline);
}

void gfx_imgpoly(SDL_Surface *surf, int n,int pt[][2],Uint8 alpha, int ax, int ay, double xu, double xv, double yu, double yv, int imgalpha, SDL_Surface *img)
{
	double dw = img->w;
	double dh = (0 - img->h);

	xu *= dw;
	xv *= dh;
	yu *= dw;
	yv *= dh;

	int sxu = xu * 65536.0;
	int sxv = xv * 65536.0;
	int syu = yu * 65536.0;
	int syv = yv * 65536.0;

	int imgsw = img->w * 65536;
	int imgsh = img->h * 65536;

	/* no alpha calculations
	 * image is in native pixel format */
	void hline_native(int x1, int x2, int y) {
		int xoff = x1 - ax;
		int yoff = y - ay;

		int u1 = sxu * xoff + syu * yoff;
		int v1 = sxv * xoff + syv * yoff;

		int len = x2-x1+1;

		Uint8 *dp = (Uint8*) surf->pixels + y * surf->pitch + x1 * surf->format->BytesPerPixel;
		Uint8 *edp = (Uint8*) dp + len * surf->format->BytesPerPixel;

		if(u1 < 0) u1 = imgsw + u1 % imgsw;
		else if(u1 >= imgsw) u1 = u1 % imgsw;
		if(v1 < 0) v1 = imgsh + v1 % imgsh;
		else if(v1 >= imgsh) v1 = v1 % imgsh;

		switch(surf->format->BytesPerPixel) {
		case 1:
			while(dp < edp) {
				Uint8 *ip = ((Uint8*) img->pixels + (v1 >> 16) * img->pitch + (u1 >> 16));

				*dp = *ip;

				dp ++;

				u1 += sxu;
				v1 += sxv;

				if(u1 < 0) u1 = imgsw + u1 % imgsw;
				else if(u1 >= imgsw) u1 = u1 % imgsw;
				if(v1 < 0) v1 = imgsh + v1 % imgsh;
				else if(v1 >= imgsh) v1 = v1 % imgsh;
			}
			break;
		case 2:
			while(dp < edp) {
				Uint16 *ip = (Uint16*) ((Uint8*) img->pixels + (v1 >> 16) * img->pitch + (u1 >> 16) * 2);

				*((Uint16 *)dp) = *ip;

				dp += 2;

				u1 += sxu;
				v1 += sxv;

				if(u1 < 0) u1 = imgsw + u1 % imgsw;
				else if(u1 >= imgsw) u1 = u1 % imgsw;
				if(v1 < 0) v1 = imgsh + v1 % imgsh;
				else if(v1 >= imgsh) v1 = v1 % imgsh;
			}
			break;
		case 3:
			while(dp < edp) {
				Uint8 *ip = ((Uint8*) img->pixels + (v1 >> 16) * img->pitch + (u1 >> 16) * 3);

				*dp++ = *ip++;
				*dp++ = *ip++;
				*dp++ = *ip++;

				u1 += sxu;
				v1 += sxv;

				if(u1 < 0) u1 = imgsw + u1 % imgsw;
				else if(u1 >= imgsw) u1 = u1 % imgsw;
				if(v1 < 0) v1 = imgsh + v1 % imgsh;
				else if(v1 >= imgsh) v1 = v1 % imgsh;
			}
			break;
		case 4:
			while(dp < edp) {
				Uint32 *ip = (Uint32*) ((Uint8*) img->pixels + (v1 >> 16) * img->pitch + (u1 >> 16) * 4);

				*((Uint32 *)dp) = *ip;

				dp += 4;

				u1 += sxu;
				v1 += sxv;

				if(u1 < 0) u1 = imgsw + u1 % imgsw;
				else if(u1 >= imgsw) u1 = u1 % imgsw;
				if(v1 < 0) v1 = imgsh + v1 % imgsh;
				else if(v1 >= imgsh) v1 = v1 % imgsh;
			}
			break;
		}
	}

	/* alpha=1, but image has alpha channel
	 * image is always in 32 bpp pixel format */
	void hline_imgalpha(int x1, int x2, int y) {
		int xoff = x1 - ax;
		int yoff = y - ay;

		int u1 = sxu * xoff + syu * yoff;
		int v1 = sxv * xoff + syv * yoff;

		int len = x2-x1+1;

		Uint8 *dp = (Uint8*) surf->pixels + y * surf->pitch + x1 * surf->format->BytesPerPixel;
		Uint8 *edp = (Uint8*) dp + len * surf->format->BytesPerPixel;

		if(u1 < 0) u1 = imgsw + u1 % imgsw;
		else if(u1 >= imgsw) u1 = u1 % imgsw;
		if(v1 < 0) v1 = imgsh + v1 % imgsh;
		else if(v1 >= imgsh) v1 = v1 % imgsh;

		switch(surf->format->BitsPerPixel) {
		case 8:
			while(dp < edp) {
				Uint32 *ip = (Uint32*) ((Uint8*) img->pixels + (v1 >> 16) * img->pitch + (u1 >> 16) * 4);
				int alpha = ((*ip >> 24) + 3) >> 5;

				Uint32 clr;
				clr = ((*ip >> 21) & 7) << 5 | // red
					  ((*ip >> 13) & 7) << 2 | // green
					  ((*ip >> 6)  & 3);  // blue

				Uint32 t1 = *dp & 0xe3;
				Uint32 t2 = *dp & 0x1c;

				*dp = (((((((clr & 0x1c) >> 2) - (t2 >> 2)) * alpha) >> 1) + t2) & 0x1c) | ((((((clr & 0xe3) - t1) * alpha) >> 3) + t1) & 0xe3);

				dp ++;

				u1 += sxu;
				v1 += sxv;

				if(u1 < 0) u1 = imgsw + u1 % imgsw;
				else if(u1 >= imgsw) u1 = u1 % imgsw;
				if(v1 < 0) v1 = imgsh + v1 % imgsh;
				else if(v1 >= imgsh) v1 = v1 % imgsh;
			}
			break;

		case 15:
			while(dp < edp) {
				Uint32 *ip = (Uint32*) ((Uint8*) img->pixels + (v1 >> 16) * img->pitch + (u1 >> 16) * 4);
				int alpha = ((*ip >> 24) + 7) >> 3;

				Uint32 clr;
				clr = ((*ip >> 19) & 31) << 10 |
					  ((*ip >> 11) & 31) << 5 |
					  ((*ip >> 3) & 31);

				Uint32 t1 = *(Uint16*)dp & 0x7c1f;
				Uint32 t2 = *(Uint16*)dp & 0x03e0;

				*((Uint16 *)dp) =
					((((((clr & 0x03e0) >> 5) - (t2 >> 5)) * alpha) + t2) & 0x03e0) |
					((((((clr & 0x7c1f) - t1) * alpha) >> 5) + t1) & 0x7c1f);

				dp += 2;

				u1 += sxu;
				v1 += sxv;

				if(u1 < 0) u1 = imgsw + u1 % imgsw;
				else if(u1 >= imgsw) u1 = u1 % imgsw;
				if(v1 < 0) v1 = imgsh + v1 % imgsh;
				else if(v1 >= imgsh) v1 = v1 % imgsh;
			}
			break;

		case 16:
			while(dp < edp) {
				Uint32 *ip = (Uint32*) ((Uint8*) img->pixels + (v1 >> 16) * img->pitch + (u1 >> 16) * 4);
				int alpha = ((*ip >> 24) + 7) >> 3;

				Uint32 clr;
				clr = ((*ip >> 19) & 31) << 11 |
					  ((*ip >> 10) & 63) << 5 |
					  ((*ip >> 3) & 31);

				Uint32 t1 = *(Uint16*)dp & 0xf81f;
				Uint32 t2 = *(Uint16*)dp & 0x07e0;

				*((Uint16 *)dp) =
					((((((clr & 0x07e0) >> 5) - (t2 >> 5)) * alpha) + t2) & 0x07e0) |
					((((((clr & 0xf81f) - t1) * alpha) >> 5) + t1) & 0xf81f);

				dp += 2;

				u1 += sxu;
				v1 += sxv;

				if(u1 < 0) u1 = imgsw + u1 % imgsw;
				else if(u1 >= imgsw) u1 = u1 % imgsw;
				if(v1 < 0) v1 = imgsh + v1 % imgsh;
				else if(v1 >= imgsh) v1 = v1 % imgsh;
			}
			break;

		case 24:
			while(dp < edp) {
				Uint32 *ip = (Uint32*) ((Uint8*) img->pixels + (v1 >> 16) * img->pitch + (u1 >> 16) * 4);
				int alpha = *ip >> 24;
				int nalpha = 255 - alpha;

				Uint32 c1, c2, c3;

				if(SDL_BYTEORDER == SDL_BIG_ENDIAN) {
					c1 = ((*ip >> 16) & 0xff) * alpha;
					c2 = ((*ip >> 8) & 0xff) * alpha;
					c3 = (*ip & 0xff) * alpha;
				} else {
					c3 = ((*ip >> 16) & 0xff) * alpha;
					c2 = ((*ip >> 8) & 0xff) * alpha;
					c1 = (*ip & 0xff) * alpha;
				}

				*dp = (*dp * nalpha + c1) >> 8; dp++;
				*dp = (*dp * nalpha + c2) >> 8; dp++;
				*dp = (*dp * nalpha + c3) >> 8; dp++;

				u1 += sxu;
				v1 += sxv;

				if(u1 < 0) u1 = imgsw + u1 % imgsw;
				else if(u1 >= imgsw) u1 = u1 % imgsw;
				if(v1 < 0) v1 = imgsh + v1 % imgsh;
				else if(v1 >= imgsh) v1 = v1 % imgsh;
			}
			break;

		case 32:
			while(dp < edp) {
				Uint32 *ip = (Uint32*) ((Uint8*) img->pixels + (v1 >> 16) * img->pitch + (u1 >> 16) * 4);
				int alpha = *ip >> 24;
#ifdef BGR
				Uint32 dc = SDL_Swap32(*(Uint32*)dp);
#else
				Uint32 dc = *(Uint32*)dp;
#endif

				Uint32 t2 = dc & 0xff00ff;
				Uint32 t1 = (*ip & 0xff00ff) - t2;
				t1 = (((t1 * alpha) >> 8) + t2) & 0xff00ff;
				dc &= 0xff00;
				t2 = (((((*ip & 0x00ff00) - dc) * alpha) >> 8) + dc) & 0xff00;

#ifdef BGR
				*((Uint32 *)dp) = SDL_Swap32(t1 | t2);
#else
				*((Uint32 *)dp) = t1 | t2;
#endif

				dp += 4;

				u1 += sxu;
				v1 += sxv;

				if(u1 < 0) u1 = imgsw + u1 % imgsw;
				else if(u1 >= imgsw) u1 = u1 % imgsw;
				if(v1 < 0) v1 = imgsh + v1 % imgsh;
				else if(v1 >= imgsh) v1 = v1 % imgsh;
			}
			break;
		}
	}

	/* alpha != 1, image has no alpha channel
	 * image is in native pixel format */
	void hline_native_alpha(int x1, int x2, int y) {
		int xoff = x1 - ax;
		int yoff = y - ay;

		int u1 = sxu * xoff + syu * yoff;
		int v1 = sxv * xoff + syv * yoff;

		int len = x2-x1+1;

		Uint8 *dp = (Uint8*) surf->pixels + y * surf->pitch + x1 * surf->format->BytesPerPixel;
		Uint8 *edp = (Uint8*) dp + len * surf->format->BytesPerPixel;

		if(u1 < 0) u1 = imgsw + u1 % imgsw;
		else if(u1 >= imgsw) u1 = u1 % imgsw;
		if(v1 < 0) v1 = imgsh + v1 % imgsh;
		else if(v1 >= imgsh) v1 = v1 % imgsh;

		Uint32 ma, cg, cng, t1, t2;

		switch(surf->format->BitsPerPixel) {
		case 8:
			ma = (alpha + 3) >> 5;

			while(dp < edp) {
				Uint8 *ip = (Uint8*) img->pixels + (v1 >> 16) * img->pitch + (u1 >> 16);

				cg = (*ip & 0x1c) >> 2;
				cng = *ip & 0xe3;
				t1 = *dp & 0xe3;
				t2 = *dp & 0x1c;
				*dp = (((((cg - (t2 >> 2)) * ma) >> 1) + t2) & 0x1c) | (((((cng - t1) * ma) >> 3) + t1) & 0xe3);

				dp ++;

				u1 += sxu;
				v1 += sxv;

				if(u1 < 0) u1 = imgsw + u1 % imgsw;
				else if(u1 >= imgsw) u1 = u1 % imgsw;
				if(v1 < 0) v1 = imgsh + v1 % imgsh;
				else if(v1 >= imgsh) v1 = v1 % imgsh;
			}
			break;

		case 15:
			ma = (alpha + 7) >> 3;

			while(dp < edp) {
				Uint16 *ip = (Uint16*) ((Uint8*) img->pixels + (v1 >> 16) * img->pitch + (u1 >> 16) * 2);

				cg = (*ip & 0x03e0) >> 5;
				cng = *ip & 0x7c1f;
				t1 = *(Uint16*)dp & 0x7c1f;
				t2 = *(Uint16*)dp & 0x03e0;
				*(Uint16*)dp = ((((cg - (t2 >> 5)) * ma) + t2) & 0x03e0) | (((((cng - t1) * ma) >> 5) + t1) & 0x7c1f);

				dp += 2;

				u1 += sxu;
				v1 += sxv;

				if(u1 < 0) u1 = imgsw + u1 % imgsw;
				else if(u1 >= imgsw) u1 = u1 % imgsw;
				if(v1 < 0) v1 = imgsh + v1 % imgsh;
				else if(v1 >= imgsh) v1 = v1 % imgsh;
			}
			break;

		case 16:
			ma = (alpha + 7) >> 3;

			while(dp < edp) {
				Uint16 *ip = (Uint16*) ((Uint8*) img->pixels + (v1 >> 16) * img->pitch + (u1 >> 16) * 2);

				cg = (*ip & 0x07e0) >> 5;
				cng = *ip & 0xf81f;
				t1 = *(Uint16*)dp & 0xf81f;
				t2 = *(Uint16*)dp & 0x07e0;
				*(Uint16*)dp = ((((cg - (t2 >> 5)) * ma) + t2) & 0x07e0) | (((((cng - t1) * ma) >> 5) + t1) & 0xf81f);

				dp += 2;

				u1 += sxu;
				v1 += sxv;

				if(u1 < 0) u1 = imgsw + u1 % imgsw;
				else if(u1 >= imgsw) u1 = u1 % imgsw;
				if(v1 < 0) v1 = imgsh + v1 % imgsh;
				else if(v1 >= imgsh) v1 = v1 % imgsh;
			}
			break;

		case 24:
			ma = 255 - alpha;

			while(dp < edp) {
				Uint8 *ip = (Uint8*) img->pixels + (v1 >> 16) * img->pitch + (u1 >> 16) * 3;

				*dp = (*dp * ma + *ip * alpha) >> 8;
				ip ++; dp ++;
				*dp = (*dp * ma + *ip * alpha) >> 8;
				ip ++; dp ++;
				*dp = (*dp * ma + *ip * alpha) >> 8;
				ip ++; dp ++;

				u1 += sxu;
				v1 += sxv;

				if(u1 < 0) u1 = imgsw + u1 % imgsw;
				else if(u1 >= imgsw) u1 = u1 % imgsw;
				if(v1 < 0) v1 = imgsh + v1 % imgsh;
				else if(v1 >= imgsh) v1 = v1 % imgsh;
			}
			break;

		case 32:
			while(dp < edp) {
				Uint32 *ip = (Uint32*) ((Uint8*) img->pixels + (v1 >> 16) * img->pitch + (u1 >> 16) * 4);

#ifdef BGR
				cg = SDL_Swap32(*ip) & 0x00ff00;
				cng = SDL_Swap32(*ip) & 0xff00ff;
#else
				cg = *ip & 0x00ff00;
				cng = *ip & 0xff00ff;
#endif

#ifdef BGR
				ma = SDL_Swap32(*(Uint32 *) dp);
#else
				ma = *(Uint32 *) dp;
#endif
				t2 = ma & 0xff00ff;
				t1 = cng - t2;
				t1 = (((t1*alpha) >> 8) + t2) & 0xff00ff;
				ma &= 0xff00;
				t2 = ((((cg - ma) * alpha) >> 8) + ma) & 0xff00;

#ifdef BGR
				*(Uint32*) dp = SDL_Swap32(t1 | t2);
#else
				*(Uint32*) dp = t1 | t2;
#endif

				dp += 4;

				u1 += sxu;
				v1 += sxv;

				if(u1 < 0) u1 = imgsw + u1 % imgsw;
				else if(u1 >= imgsw) u1 = u1 % imgsw;
				if(v1 < 0) v1 = imgsh + v1 % imgsh;
				else if(v1 >= imgsh) v1 = v1 % imgsh;
			}
			break;
		}
	}

	/* alpha != 1, image has alpha channel
	 * image is in 32bpp pixel format */
	void hline_multalpha(int x1, int x2, int y) {
		int xoff = x1 - ax;
		int yoff = y - ay;

		int u1 = sxu * xoff + syu * yoff;
		int v1 = sxv * xoff + syv * yoff;

		int len = x2-x1+1;

		Uint8 *dp = (Uint8*) surf->pixels + y * surf->pitch + x1 * surf->format->BytesPerPixel;
		Uint8 *edp = (Uint8*) dp + len * surf->format->BytesPerPixel;

		if(u1 < 0) u1 = imgsw + u1 % imgsw;
		else if(u1 >= imgsw) u1 = u1 % imgsw;
		if(v1 < 0) v1 = imgsh + v1 % imgsh;
		else if(v1 >= imgsh) v1 = v1 % imgsh;

		switch(surf->format->BitsPerPixel) {
		case 8:
			while(dp < edp) {
				Uint32 *ip = (Uint32*) ((Uint8*) img->pixels + (v1 >> 16) * img->pitch + (u1 >> 16) * 4);
				int ralpha = ((((*ip >> 24) * alpha) >> 8) + 3) >> 5;

				Uint32 clr;
				clr = ((*ip >> 21) & 7) << 5 | // red
					  ((*ip >> 13) & 7) << 2 | // green
					  ((*ip >> 6)  & 3);  // blue

				Uint32 t1 = *dp & 0xe3;
				Uint32 t2 = *dp & 0x1c;

				*dp = (((((((clr & 0x1c) >> 2) - (t2 >> 2)) * ralpha) >> 1) + t2) & 0x1c) | ((((((clr & 0xe3) - t1) * ralpha) >> 3) + t1) & 0xe3);

				dp ++;

				u1 += sxu;
				v1 += sxv;

				if(u1 < 0) u1 = imgsw + u1 % imgsw;
				else if(u1 >= imgsw) u1 = u1 % imgsw;
				if(v1 < 0) v1 = imgsh + v1 % imgsh;
				else if(v1 >= imgsh) v1 = v1 % imgsh;
			}
			break;

		case 15:
			while(dp < edp) {
				Uint32 *ip = (Uint32*) ((Uint8*) img->pixels + (v1 >> 16) * img->pitch + (u1 >> 16) * 4);
				int ralpha = ((((*ip >> 24) * alpha) >> 8) + 7) >> 3;

				Uint32 clr;
				clr = ((*ip >> 19) & 31) << 10 |
					  ((*ip >> 11) & 31) << 5 |
					  ((*ip >> 3) & 31);

				Uint32 t1 = *(Uint16*)dp & 0x7c1f;
				Uint32 t2 = *(Uint16*)dp & 0x03e0;

				*((Uint16 *)dp) =
					((((((clr & 0x03e0) >> 5) - (t2 >> 5)) * ralpha) + t2) & 0x03e0) |
					((((((clr & 0x7c1f) - t1) * ralpha) >> 5) + t1) & 0x7c1f);

				dp += 2;

				u1 += sxu;
				v1 += sxv;

				if(u1 < 0) u1 = imgsw + u1 % imgsw;
				else if(u1 >= imgsw) u1 = u1 % imgsw;
				if(v1 < 0) v1 = imgsh + v1 % imgsh;
				else if(v1 >= imgsh) v1 = v1 % imgsh;
			}
			break;

		case 16:
			while(dp < edp) {
				Uint32 *ip = (Uint32*) ((Uint8*) img->pixels + (v1 >> 16) * img->pitch + (u1 >> 16) * 4);
				int ralpha = ((((*ip >> 24) * alpha) >> 8) + 7) >> 3;

				Uint32 clr;
				clr = ((*ip >> 19) & 31) << 11 |
					  ((*ip >> 10) & 63) << 5 |
					  ((*ip >> 3) & 31);

				Uint32 t1 = *(Uint16*)dp & 0xf81f;
				Uint32 t2 = *(Uint16*)dp & 0x07e0;

				*((Uint16 *)dp) =
					((((((clr & 0x07e0) >> 5) - (t2 >> 5)) * ralpha) + t2) & 0x07e0) |
					((((((clr & 0xf81f) - t1) * ralpha) >> 5) + t1) & 0xf81f);

				dp += 2;

				u1 += sxu;
				v1 += sxv;

				if(u1 < 0) u1 = imgsw + u1 % imgsw;
				else if(u1 >= imgsw) u1 = u1 % imgsw;
				if(v1 < 0) v1 = imgsh + v1 % imgsh;
				else if(v1 >= imgsh) v1 = v1 % imgsh;
			}
			break;

		case 24:
			while(dp < edp) {
				Uint32 *ip = (Uint32*) ((Uint8*) img->pixels + (v1 >> 16) * img->pitch + (u1 >> 16) * 4);
				int ralpha = ((*ip >> 24) * alpha) >> 8;
				int nalpha = 255 - ralpha;

				Uint32 c1, c2, c3;

				if(SDL_BYTEORDER == SDL_BIG_ENDIAN) {
					c1 = ((*ip >> 16) & 0xff) * ralpha;
					c2 = ((*ip >> 8) & 0xff) * ralpha;
					c3 = (*ip & 0xff) * ralpha;
				} else {
					c3 = ((*ip >> 16) & 0xff) * ralpha;
					c2 = ((*ip >> 8) & 0xff) * ralpha;
					c1 = (*ip & 0xff) * ralpha;
				}

				*dp = (*dp * nalpha + c1) >> 8; dp++;
				*dp = (*dp * nalpha + c2) >> 8; dp++;
				*dp = (*dp * nalpha + c3) >> 8; dp++;

				u1 += sxu;
				v1 += sxv;

				if(u1 < 0) u1 = imgsw + u1 % imgsw;
				else if(u1 >= imgsw) u1 = u1 % imgsw;
				if(v1 < 0) v1 = imgsh + v1 % imgsh;
				else if(v1 >= imgsh) v1 = v1 % imgsh;
			}
			break;

		case 32:
			while(dp < edp) {
				Uint32 *ip = (Uint32*) ((Uint8*) img->pixels + (v1 >> 16) * img->pitch + (u1 >> 16) * 4);
				int ralpha = ((*ip >> 24) * alpha) >> 8;
#ifdef BGR
				Uint32 dc = SDL_Swap32(*(Uint32*)dp);
#else
				Uint32 dc = *(Uint32*)dp;
#endif

				Uint32 t2 = dc & 0xff00ff;
				Uint32 t1 = (*ip & 0xff00ff) - t2;
				t1 = (((t1 * ralpha) >> 8) + t2) & 0xff00ff;
				dc &= 0xff00;
				t2 = (((((*ip & 0x00ff00) - dc) * ralpha) >> 8) + dc) & 0xff00;

#ifdef BGR
				*((Uint32 *)dp) = SDL_Swap32(t1 | t2);
#else
				*((Uint32 *)dp) = t1 | t2;
#endif

				dp += 4;

				u1 += sxu;
				v1 += sxv;

				if(u1 < 0) u1 = imgsw + u1 % imgsw;
				else if(u1 >= imgsw) u1 = u1 % imgsw;
				if(v1 < 0) v1 = imgsh + v1 % imgsh;
				else if(v1 >= imgsh) v1 = v1 % imgsh;
			}
			break;
		}
	}

	if(alpha == 255) {
		if(imgalpha)
			scanpoly(surf->clip_rect.x, surf->clip_rect.x + surf->clip_rect.w - 1, surf->clip_rect.y, surf->clip_rect.y + surf->clip_rect.h - 1, n, pt, hline_imgalpha);
		else
			scanpoly(surf->clip_rect.x, surf->clip_rect.x + surf->clip_rect.w - 1, surf->clip_rect.y, surf->clip_rect.y + surf->clip_rect.h - 1, n, pt, hline_native);
	} else {
		if(imgalpha) {
			scanpoly(surf->clip_rect.x, surf->clip_rect.x + surf->clip_rect.w - 1, surf->clip_rect.y, surf->clip_rect.y + surf->clip_rect.h - 1, n, pt, hline_multalpha);
		} else {
			scanpoly(surf->clip_rect.x, surf->clip_rect.x + surf->clip_rect.w - 1, surf->clip_rect.y, surf->clip_rect.y + surf->clip_rect.h - 1, n, pt, hline_native_alpha);
		}
	}
}

/* multalpha blit scanline: 8 bit */
static inline void gfx_multalpha_scan_8(Uint32 *ip, Uint8 *dp, int w, Uint8 alpha)
{
	int i, ralpha;
	Uint32 t1, t2, clr;

	for(i=0; i<w; i++) {
		ralpha = ((((*ip >> 24) * alpha) >> 8) + 3) >> 5;
		/*
		nalpha = 256 - ralpha;

		Uint32 r = (((*dp & 0xe0) * nalpha + ((*ip >> 16) & 0xff) * ralpha) >> 8) & 0xff;
		Uint32 g = ((((*dp & 0x1c) << 3) * nalpha + ((*ip >> 8) & 0xff) * ralpha) >> 8) & 0xff;
		Uint32 b = ((((*dp & 3) << 6) * nalpha + (*ip & 0xff) * ralpha) >> 8) & 0xff;

		*dp = (r & 0xe0) | ((g & 0xe0) >> 3) | ((b & 0xc0) >> 6);
		*/

		clr = ((*ip >> 21) & 7) << 5 | // red
		      ((*ip >> 13) & 7) << 2 | // green
		      ((*ip >> 6)  & 3);  // blue

		t1 = *dp & 0xe3;
		t2 = *dp & 0x1c;

		*dp = (((((((clr & 0x1c) >> 2) - (t2 >> 2)) * ralpha) >> 1) + t2) & 0x1c) | ((((((clr & 0xe3) - t1) * ralpha) >> 3) + t1) & 0xe3);

		ip ++;
		dp ++;
	}
}

/* multalpha blit scanline: 15 bit */
static inline void gfx_multalpha_scan_15(Uint32 *ip, Uint16 *dp, int w, Uint8 alpha)
{
	int i, ralpha;
	Uint32 t1, t2, clr;

	for(i=0; i<w; i++) {
		ralpha = ((((*ip >> 24) * alpha) >> 8) + 7) >> 3;

		clr = ((*ip >> 19) & 31) << 10 |
		      ((*ip >> 11) & 31) << 5 |
		      ((*ip >> 3) & 31);

		t1 = *dp & 0x7c1f;
		t2 = *dp & 0x03e0;

		*dp = ((((((clr & 0x03e0) >> 5) - (t2 >> 5)) * ralpha) + t2) & 0x03e0) |
			((((((clr & 0x7c1f) - t1) * ralpha) >> 5) + t1) & 0x7c1f);

		ip ++;
		dp ++;
	}
}

/* multalpha blit scanline: 16 bit */
static inline void gfx_multalpha_scan_16(Uint32 *ip, Uint16 *dp, int w, Uint8 alpha)
{
	int i, ralpha;
	Uint32 t1, t2, clr;

	for(i=0; i<w; i++) {
		ralpha = ((((*ip >> 24) * alpha) >> 8) + 7) >> 3;

		clr = ((*ip >> 19) & 31) << 11 |
		      ((*ip >> 10) & 63) << 5 |
		      ((*ip >> 3) & 31);

		t1 = *dp & 0xf81f;
		t2 = *dp & 0x07e0;

		*dp = ((((((clr & 0x07e0) >> 5) - (t2 >> 5)) * ralpha) + t2) & 0x07e0) |
			((((((clr & 0xf81f) - t1) * ralpha) >> 5) + t1) & 0xf81f);

		ip ++;
		dp ++;
	}
}

/* multalpha blit scanline: 24 bit */
static inline void gfx_multalpha_scan_24(Uint32 *ip, Uint8 *dp, int w, Uint8 alpha)
{
	int i, ralpha;
	Uint32 c1, c2, c3;
	Uint32 nalpha;

	for(i=0; i<w; i++) {
		ralpha = ((*ip >> 24) * alpha) >> 8;
		nalpha = 255 - ralpha;

		if(SDL_BYTEORDER == SDL_BIG_ENDIAN) {
			c1 = ((*ip >> 16) & 0xff) * ralpha;
			c2 = ((*ip >> 8) & 0xff) * ralpha;
			c3 = (*ip & 0xff) * ralpha;
		} else {
			c3 = ((*ip >> 16) & 0xff) * ralpha;
			c2 = ((*ip >> 8) & 0xff) * ralpha;
			c1 = (*ip & 0xff) * ralpha;
		}

		*dp = (*dp * nalpha + c1) >> 8; dp++;
		*dp = (*dp * nalpha + c2) >> 8; dp++;
		*dp = (*dp * nalpha + c3) >> 8; dp++;

		ip++;
	}
}

/* multalpha blit scanline: 32 bit */
static inline void gfx_multalpha_scan_32(Uint32 *ip, Uint32 *dp, int w, Uint8 alpha)
{
	int i, ralpha;
	Uint32 t1, t2, dc;

	for(i=0; i<w; i++) {
		ralpha = ((*ip >> 24) * alpha) >> 8;

#ifdef BGR
		dc = SDL_Swap32(*dp);
#else
		dc = *dp;
#endif

		t2 = dc & 0xff00ff;
		t1 = (*ip & 0xff00ff) - t2;
		t1 = (((t1 * ralpha) >> 8) + t2) & 0xff00ff;
		dc &= 0xff00;
		t2 = (((((*ip & 0x00ff00) - dc) * ralpha) >> 8) + dc) & 0xff00;

#ifdef BGR
		*dp = SDL_Swap32(t1 | t2);
#else
		*dp = t1 | t2;
#endif

		ip ++;
		dp ++;
	}
}

/* multalpha blit (pixel*given) */
void gfx_multalpha_blit(SDL_Surface *src, SDL_Surface *dst, int x, int y, Uint8 alpha)
{
	int ix, iy; /* image start offset */
	int dx, dy; /* destination start offset */
	int w, h;   /* width/height */

	assert(src->format->BitsPerPixel == 32);

	w = src->w;
	h = src->h;

	ix = 0;
	iy = 0;

	dx = x;
	dy = y;

	/* do some clipping */

	if(dx < dst->clip_rect.x) {
		int off = dst->clip_rect.x - dx;

		w -= off;
		ix += off;
		dx = dst->clip_rect.x;
	}

	if(dy < dst->clip_rect.y) {
		int off = dst->clip_rect.y - dy;

		h -= off;
		iy += off;
		dy = dst->clip_rect.y;
	}

	if(dx + w > dst->clip_rect.x + dst->clip_rect.w) {
		int off = dx + w - (dst->clip_rect.x + dst->clip_rect.w);

		w -= off;
	}

	if(dy + h > dst->clip_rect.y + dst->clip_rect.h) {
		int off = dy + h - (dst->clip_rect.y + dst->clip_rect.h);

		h -= off;
	}

	if(w < 1 || h < 1) return;

	/* first pixel in source image */
	Uint8 *ip = ((Uint8*) src->pixels) + iy * src->pitch + ix * 4;
	/* first pixel in destination */
	Uint8 *dp = ((Uint8*) dst->pixels) + dy * dst->pitch + dx * dst->format->BytesPerPixel;

	int i;
	for(i=0; i<h; i++) {
		switch(dst->format->BitsPerPixel) {
		case 8:
			gfx_multalpha_scan_8((Uint32 *) ip, dp, w, alpha);
			break;
		case 15:
			gfx_multalpha_scan_15((Uint32 *) ip, (Uint16 *) dp, w, alpha);
			break;
		case 16:
			gfx_multalpha_scan_16((Uint32 *) ip, (Uint16 *) dp, w, alpha);
			break;
		case 24:
			gfx_multalpha_scan_24((Uint32 *) ip, (Uint8 *) dp, w, alpha);
			break;
		case 32:
			gfx_multalpha_scan_32((Uint32 *) ip, (Uint32 *) dp, w, alpha);
			break;
		}

		ip += src->pitch;
		dp += dst->pitch;
	}
}
