%{
	#include "cmd.h"
	#include "parse.tab.h"

	#define MAX_STR_CONST 1024
	char string_buf[MAX_STR_CONST+1];
	char *string_buf_ptr;

	#define YY_INPUT(buf, result, max_size) \
		{ \
			result = doinput(buf, max_size); \
		}
	
	int yyline = 1;
%}

%x str
%x incl

%%

[0-9]*\.[0-9]+e[+-][0-9]+	yylval.d = atof(yytext); return NUMBER;
[-+][0-9]*\.[0-9]+e[+-][0-9]+	yylval.d = atof(yytext); return NUMBER;
[0-9]*\.[0-9]+e[0-9]+	yylval.d = atof(yytext); return NUMBER;
[-+][0-9]*\.[0-9]+e[0-9]+	yylval.d = atof(yytext); return NUMBER;
[0-9]+e[+-][0-9]+	yylval.d = atof(yytext); return NUMBER;
[-+][0-9]+e[+-][0-9]+	yylval.d = atof(yytext); return NUMBER;
[0-9]*\.[0-9]+	yylval.d = atof(yytext); return NUMBER;
[-+][0-9]*\.[0-9]+	yylval.d = atof(yytext); return NUMBER;
[0-9]+	yylval.d = atof(yytext); return NUMBER;
[-+][0-9]+	yylval.d = atof(yytext); return NUMBER;
0x[0-9a-fA-F]{4} yylval.i = strtol(yytext, NULL, 0); return HEX;

dot         return DOT;
line        return LINE;
arrow       return ARROW;
rect        return RECT;
poly        return POLY;
apoly       return APOLY;

text        return TEXT;

lines       return LINES;
\.          return '.';

left        return LEFT;
center      return CENTER;
right       return RIGHT;
bottom      return BOTTOM;
top         return TOP;
circle      return CIRCLE;
fillcircle  return FILLCIRCLE;
fillrect    return FILLRECT;

alpha       return ALPHA;
colormap    return COLORMAP;
color       return COLOR;
bg          return BG;
thick       return THICK;
dash        return DASH;

push        return PUSH;
pop         return POP;
mult        return MULT;
rotate      return ROTATE;
scale       return SCALE;
shift       return SHIFT;
viewport    return VIEWPORT;
ident       return IDENT;

events      return EVENTS;
aspect      return ASPECT;
on          return ON;
off         return OFF;
area        return AREA;
delarea     return DELAREA;
killareas   return KILLAREAS;

frame       return FRAME;
flushdelay  return FLUSHDELAY;
flush       return FLUSH;
screenshot  return SCREENSHOT;
screendump  return SCREENDUMP;

sleep       return SLEEP;

keepalive   return KEEPALIVE;
quit        return QUIT;
echo        return ECHOCMD;
echopt      return ECHOPTCMD;
title       return TITLE;

resize      return RESIZE;
warp        return WARP;

macro       return MACRO;
endmacro    return ENDMACRO;

forget      return FORGET;

invoke      return INVOKE;

image       return IMAGE;
image_set   return IMAGE_SET;
blit        return BLIT;
imgpoly     return IMGPOLY;
imgapoly    return IMGAPOLY;

defcursor   return DEFCURSOR;
setcursor   return SETCURSOR;

clip        return CLIP;

[a-zA-Z_][a-zA-Z_0-9]*	yylval.s = strdup(yytext); return SYMBOL;
'.'         yylval.i = yytext[1]; return CHARACTER;

#[a-fA-F0-9]{6} yylval.s = yytext; return COLORHEX;
#[a-fA-F0-9]{3} yylval.s = yytext; return COLORHEX;
#[a-fA-F0-9]{2} yylval.s = yytext; return COLORHEX;
#[a-fA-F0-9]{1} yylval.s = yytext; return COLORHEX;

!.*[\n] yyline ++;
[\n]    yyline ++;
[ \t\r]+

\"      string_buf_ptr = string_buf; BEGIN(str);

[()-] return yytext[0];

.	yyerror("Input unrecognized");

<str>\"        { /* saw closing quote - all done */
        BEGIN(INITIAL);
        *string_buf_ptr = '\0';
		yylval.s = string_buf;
		return SYMBOL;
        }

<str>\\[0-7]{1,3} {
        /* octal escape sequence */
        int result;

        (void) sscanf( yytext + 1, "%o", &result );

        if ( result > 0xff ) {
			yyerror("Invalid octal escape sequence - constant is out of bounds");
			yyterminate();
		}

        *string_buf_ptr++ = result;
        }

<str>\\[0-9]+ {
		yyerror("Invalid escape sequence\n");
        }

<str>\\n  *string_buf_ptr++ = '\n';
<str>\\t  *string_buf_ptr++ = '\t';
<str>\\r  *string_buf_ptr++ = '\r';
<str>\\b  *string_buf_ptr++ = '\b';
<str>\\f  *string_buf_ptr++ = '\f';

<str>\\.  *string_buf_ptr++ = yytext[1];
<str>[\n]  { *string_buf_ptr++ = yytext[0]; yyline ++; }

<str>[^\\\n\"]+        {
        char *yptr = yytext;

        while ( *yptr && (string_buf_ptr - string_buf) < MAX_STR_CONST )
                *string_buf_ptr++ = *yptr++;
        }

